"""pruning criterion."""
# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from intel_extension_for_transformers.optimization.utils.utility import LazyImport
torch = LazyImport('torch')


CRITERIA = {}


def register_criterion(name):
    """Register a criterion to the registry."""

    def register(criterion):
        CRITERIA[name] = criterion
        return criterion

    return register


def get_criterion(config, modules):
    """Get registered criterion class."""
    name = config["criterion_type"]
    if name not in CRITERIA.keys():  # pragma: no cover
        assert False, f"criteria does not support {name}, currently only support {CRITERIA.keys()}"
    return CRITERIA[name](modules, config)


class PruningCriterion:
    """Pruning base criterion.

    Args:
        config: A config dict object that includes information about pruner and pruning criterion.
        modules: A dict {"module_name": Tensor} that stores the pruning modules' weights.
    
    Attributes:
        scores: A dict {"module_name": Tensor} that stores the scores of pruning modules.
    """

    def __init__(self, modules, config):
        """Initiliaze a pruning criterion."""
        self.scores = {}
        self.modules = modules
        self.config = config

    def on_step_begin(self):
        """Calculate and store the pruning scores of pruning modules at the beginning of a step."""
        pass

    def on_before_optimizer_step(self):
        """Calculate and store the pruning scores of pruning modules before the optimizer step."""
        pass

    def on_after_optimizer_step(self):  # pragma: no cover
        """Calculate and store the pruning scores of pruning modules after the optimizer step."""
        pass


@register_criterion('magnitude')
class MagnitudeCriterion(PruningCriterion):
    """Pruning criterion.
    
    The magnitude criterion_class is derived from PruningCriterion.
    The magnitude value is used to score and determine if a weight is to be pruned.

    Args:
        config: A config dict object that includes information about pruner and pruning criterion.
        modules: A dict {"module_name": Tensor} that stores the pruning modules' weights.
    
    Attributes:
        scores: A dict {"module_name": Tensor} that stores the scores of pruning modules.
    """

    def __init__(self, modules, config):
        """Initiliaze a magnitude pruning criterion."""
        super(MagnitudeCriterion, self).__init__(modules, config)

    def on_step_begin(self):
        """Calculate and store the pruning scores based on magtinude criterion."""
        with torch.no_grad():
            for key in self.modules.keys():
                p = self.modules[key].weight.data
                self.scores[key] = p


@register_criterion('gradient')
class GradientCriterion(PruningCriterion):
    """Pruning criterion.
    
    The gradient criterion_class is derived from PruningCriterion.
    The absolute value of gradient is used to score and determine if a weight is to be pruned.

    Args:
        config: A config dict object that includes information about pruner and pruning criterion.
        modules: A dict {"module_name": Tensor} that stores the pruning modules' weights.
    
    Attributes:
        scores: A dict {"module_name": Tensor} that stores the scores of pruning modules.
    """

    def __init__(self, modules, config):
        """Initiliaze a gradient pruning criterion."""
        super(GradientCriterion, self).__init__(modules, config)
        assert self.config.end_step > 0, "please set end_step > 0 for gradient based criterion"

    def on_before_optimizer_step(self):
        """Calculate and store the pruning scores based on gradient criterion."""
        with torch.no_grad():
            for key in self.modules.keys():
                p = self.modules[key].weight
                self.scores[key] = torch.abs(p.grad)


@register_criterion('snip')
class SnipCriterion(PruningCriterion):
    """Pruning criterion.
    
    The snip criterion_class is derived from PruningCriterion.
    The product of magnitude and gradient is used to score and determine if a weight is to be pruned.
    Please refer to SNIP: Single-shot Network Pruning based on Connection Sensitivity.
    (https://arxiv.org/abs/1810.02340)

    Args:
        config: A config dict object that includes information about pruner and pruning criterion.
        modules: A dict {"module_name": Tensor} that stores the pruning modules' weights.
    
    Attributes:
        scores: A dict {"module_name": Tensor} that stores the scores of pruning modules.
    """

    def __init__(self, modules, config):
        """Initiliaze a snip pruning criterion."""
        super(SnipCriterion, self).__init__(modules, config)
        assert self.config.end_step > 0, "please set end_step > 0 for gradient based criterion"

    def on_before_optimizer_step(self):
        """Calculate and store the pruning scores based on snip criterion."""
        ##self.mask_weights()
        with torch.no_grad():
            for key in self.modules.keys():
                p = self.modules[key].weight
                self.scores[key] = torch.abs(p * p.grad)


@register_criterion('snip_momentum')
class SnipMomentumCriterion(PruningCriterion):
    """Pruning criterion.
    
    The snip_momentum criterion_class is derived from PruningCriterion.
    A momentum mechanism is used to calculate snip score, which determines if a weight is to be pruned.

    Args:
        config: A config dict object that includes information about pruner and pruning criterion.
        modules: A dict {"module_name": Tensor} that stores the pruning modules' weights.
        alpha: A parameter that determines how much of the snip score is preserved from last pruning step.
        beta: A parameter that determines how much of the snip score is updated at the current step.
    
    Attributes:
        scores: A dict {"module_name": Tensor} that stores the scores of pruning modules.
    """

    def __init__(self, modules, config):
        """Initiliaze a snip_momentum pruning criterion."""
        super(SnipMomentumCriterion, self).__init__(modules, config)
        assert self.config.end_step > 0, "please set end_step > 0 for gradient based criterion"
        for key in modules.keys():
            p = modules[key].weight
            self.scores[key] = torch.zeros(p.shape).to(p.device)

        self.alpha = 0.9
        self.beta = 1.0

    def on_before_optimizer_step(self):
        """Calculate and store the pruning scores based on snip_momentum criterion."""
        with torch.no_grad():
            for key in self.modules.keys():
                p = self.modules[key].weight
                self.scores[key] *= self.alpha
                self.scores[key] += self.beta * torch.abs(p * p.grad)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The supported pattern file."""

from abc import abstractmethod

# for complicated pattern, use a several lists and tuple to represent the main computation flow
# and sub-graphs
# for example, layer_norm pattern in bert_large:
# pattern: [ [(0, 'Mean'), (1, 'SquaredDifference'), (2, 'Mean'), (3, 'AddV2'), (4, 'Rsqrt'),
#             (5, 'Mul'),
#             (7 ,'Mul'), (8, 'Sub'), (9, 'AddV2')], [(5, 'Mul'), (6, 'Mul'), (9, 'AddV2')] ]
# the pattern has order, but need connections check cause of the staggered edges
# for detailed info, please see search_pattern func in compile/graph_utils.py
# supported patterns should have order
supported_patterns = [
    'InputFile',
    'InputData',
    'RemoveConstantOP',
    'CollectQuantInfo',
    'TorchInnerProductInsertBias',
    'LayerNorm',
    'WordEmbeddings',
    'MergedEmbeddingbag',
    'EmbeddingBag',
    'TokenTypeEmbeddings',
    'TokenTypeEmbeddingsV1',
    'PositionEmbeddings',
    'PositionEmbeddingsV1',
    'GenerateSequence',
    'PaddingSequence',
    'AttentionMaskLengthAdaptiveExpandIndices',
    'MatMulWithBias',
    'LastLayerShape',
    'InteractFeatures',
    'AttentionReshape',

    # Text Encoder
    'TextEncoder_WordEmbedding',
    'TextEncoder_QReshape',            
    'TextEncoder_KVReshape',
    'TextEncoder_AttentionMaskAddReshape',
    'TextEncoder_SoftmaxReshape',
    'TextEncoder_MulReshape',
    'TextEncoder_AttentionReshape',
    'TextEncoder_CasualAttentionMask',

    #vae decoder & Transformer2Dmodel
    'AttentionBlock_Resize2Gather',
    'GroupNorm',
    'AttentionBlock_QKVPreReshape',
    'AttentionBlock_AttentionMaskAddReshape',
    'AttentionBlock_ConstantOfShapeWithMul',

    'Transformer2Dmodel_GetSampleBatch',
    'Transformer2Dmodel_SampleSlice',
    'Transformer2Dmodel_EncoderHiddenStatesReshape',
    'Transformer2Dmodel_ConstantOfShapeWithMul',
    'Transformer2Dmodel_QKVPreReshape',
    'Transformer2Dmodel_QKVReshape',
    'AttentionBlock_QKVReshape',
    'Transformer2Dmodel_QKVReshapeTo4D',
    'Transformer2Dmodel_AttentionMaskAddReshape',
    'Transformer2Dmodel_FFNInputSlice',
    'Transformer2Dmodel_FFNInputSlice_1',
    'Transformer2DModel_UpBlockResize',

    'QKVReshape',
    'DecoderAttnReshape',
    'ConvReshape',
    'AddClsToken',
    'TransposeBatchMatMul',
    'Gelu',
    'MatMulWithBiasGelu',
    'MatMulWithBiasAdd',
    'AddEmbeddings',
    'MatMulWithBiasTanh',
    'MatMulWithBiasRelu',
    'MatMulWithBiasSigmoid',
    "RestoreHiddenStatesInLengthAdaptiveUpdateIndices",
    "AttentionOutputLayerNormLengthAdaptiveExpandIndices",
    "ReshapeBeforeAndAfterAttentionOutLayerNormGatherElements",
    "ReshapeBeforeRestoreHiddenStates",
    "ReshapeAfterRestoreHiddenStates",
    'LayerNormWithReduceMean',
    'LayerNormWithTranspose',
    'StartEndLogits',
    'CastTo',

    # GPT-J
    'LowerAllTuples',
    'TorchEmbedding',
    'InnerproductReshapeFusion',
    'MatMulWithTranspose',
    'InnerproductWithBiasGelu',
    'SliceMask',
    'ArangewithReciprocal',
    'InnerproductwithSlice',
    'RoraryPosEmb',
    'EinsumwithArange',
    'RemoveSlice',
    'RemoveRange',
    'RemoveLastView',
    # 'RemoveDuplicatedQuantNode',
    
    'InsertQuantNode',
    'InsertBF16Node',
    'TorchInsertBF16Node',
    'QunatizeFusion',
    'QKVMerge',
    'ReshapeFusion',
    'StableDiffusion_bf16Convert',
    'StableDiffusion_ReshapeFusion',
    'OperatorAdaptor',

    'MultiHeadAttention',
    'OutputData',
    'QuantizedGraphDtypeRefactor',
    'Int8BF16MixedPrecisionChecker',
]

# for superbert, superbert patterns are huge patterns based on supported patterns
superbert_patterns = []

PATTERNS = {}


def pattern_registry(pattern_type):
    """The class decorator used to register all Algorithm subclasses.

    Args:
        cls (class): The class of register.
        pattern_type (str): The pattern registration name

    Returns:
        cls: The class of register.
    """
    def decorator_pattern(cls):
        """The pattern decorator."""
        if pattern_type in PATTERNS:
            raise ValueError('Cannot have two patterns with the same name')
        PATTERNS[pattern_type] = cls
        return cls

    return decorator_pattern


class Pattern(object):
    """The bass pattern class."""
    @abstractmethod
    def __call__(self, model, *args, **kwargs):
        """The __call__ function of the pattern class."""
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The InsertBF16Node Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
from ..ops import Tensor
from .subgraph_matcher import EXECUTOR_TYPE
import numpy as np


@pattern_registry(pattern_type='InsertBF16Node')
class InsertBF16Node(Pattern):
    """The InsertBF16Node pattern.

    Fuse the original sub-graph into the custom acceleration 'InsertBF16Node' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """

    def __call__(self, model):
        """The __call__ function of this pattern class."""
        def fp32_to_bf16(fp32_np):
            assert(fp32_np.dtype==np.float32)
            int32_np = fp32_np.view(dtype=np.int32)
            int32_np = int32_np >> 16
            bf16_np = int32_np.astype(np.uint16)
            return bf16_np

        for node in model.nodes:
            if node.op_type in EXECUTOR_TYPE and \
              EXECUTOR_TYPE[node.op_type] == "InnerProduct" and \
              node.input_tensors[1].dtype == "fp16":
                node.input_tensors[1].dtype = 'bf16'
                if len(node.input_tensors) > 2:
                    bias_fp32 = node.input_tensors[2].data
                    if node.input_tensors[2].dtype == 'fp32':
                        node.input_tensors[2].data = fp32_to_bf16(bias_fp32)
                    node.input_tensors[2].dtype = 'bf16'
                input_tensor = node.input_tensors[0]
                input_name = input_tensor.name
                quant_output = Tensor(name=input_name + "_quant",
                                      source_op=[node.name + "_quant"],
                                      dest_op=[node.name],
                                      dtype='bf16')
                quantize_op = util.construct_node(
                    node_name=node.name + "_quant",
                    op_type='Quantize',
                    input_tensors=[input_tensor],
                    output_tensors=[quant_output],
                    attr=OrderedDict({'output_dtype': 'bf16'}))
                insert_idx = model.get_node_id(node.name)
                model.insert_nodes(insert_idx, [quantize_op])
                node.input_tensors[0] = quant_output
                dest_op = node.output_tensors[0].dest_op
                if dest_op and model.get_node_by_name(dest_op[0]).op_type != 'LayerNorm':
                    next_node = model.get_node_by_name(dest_op[0])
                    next_dest_op = next_node.output_tensors[0].dest_op
                    # when next_dest_op is Ouput op, output_type should be fp32
                    if not next_dest_op:
                        continue
                    node.attr['output_dtype'] = 'bf16'


        remove_duplicate_set = set()
        duplicate_list=[]
        for node in model.nodes:
            sz = len(remove_duplicate_set)
            remove_duplicate_set.add(node.output_tensors[0].name)
            new_sz = len(remove_duplicate_set)
            if new_sz == sz:
                duplicate_list.append(node.name)
        model.remove_nodes(duplicate_list)

        if duplicate_list:
            for node in model.nodes:
                next_op = node.output_tensors[0].dest_op
                if node.op_type in EXECUTOR_TYPE and \
                  EXECUTOR_TYPE[node.op_type] == "Matmul" and \
                  model.get_node_by_name(next_op[0]).op_type == 'Softmax':
                  node.attr['output_dtype'] = 'bf16'
                  model.get_node_by_name(next_op[0]).attr['output_dtype'] = 'bf16'

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The Transformer2Dmodel_EncoderHiddenStatesReshape pattern."""

from .pattern import Pattern, pattern_registry
from collections import OrderedDict
from .. import graph_utils as util
from ..ops import Tensor
from .. import logger


@pattern_registry(pattern_type='Transformer2Dmodel_EncoderHiddenStatesReshape')
class Transformer2Dmodel_EncoderHiddenStatesReshape(Pattern):
    """The Transformer2Dmodel_EncoderHiddenStatesReshape pattern.

    Fuse the original sub-graph into the custom acceleration 'Transformer2Dmodel_EncoderHiddenStatesReshape' graph.
    The search strategy is based on the following pattern mapping configs for the stable textEncoderV1-5.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'Transformer2Dmodel_EncoderHiddenStatesReshape': [
                {
                    'patterns': {
                        'in': [[(0, 'Input'), (1, 'MatMul')]],
                    },
                },
            ]
        }

        pattern = pattern_mapping_config['Transformer2Dmodel_EncoderHiddenStatesReshape'][0]['patterns']['in']
        patterns_nodes_name = util.search_pattern(pattern, model)
        logger.info('Transformer2Dmodel_EncoderHiddenStatesReshape mathched...')
        logger.debug('Transformer2Dmodel_EncoderHiddenStatesReshape = {}'.format(patterns_nodes_name))
        if len(patterns_nodes_name) != 0:
            first_matmul_node_idx = -1
            all_dest_op = []
            for i in range(len(patterns_nodes_name)):
                input_node = model.get_node_by_name(patterns_nodes_name[i][0])
                encoder_hidden_states_tensor = input_node.output_tensors[2]

                matmul_node_name = patterns_nodes_name[i][1]
                matmul_node = model.get_node_by_name(matmul_node_name)
                matmul_node_idx = model.get_node_id(matmul_node.name)
                if i == 0:
                    first_matmul_node_idx = matmul_node_idx

                #create a new 2d reshape node
                new_node_name = 'encoder_hidden_states/reshape_2d'
                new_node_output_tensor_name = encoder_hidden_states_tensor.name + '_2d'

                input_tensors = [encoder_hidden_states_tensor]
                output_tensor = [
                    Tensor(name=new_node_output_tensor_name,
                           source_op=[new_node_name],
                           dest_op=[],
                           dtype=matmul_node.output_tensors[0].dtype)
                ]
                new_node = util.construct_node(node_name=new_node_name,
                                               op_type='Reshape',
                                               input_tensors=input_tensors,
                                               output_tensors=output_tensor)
                attr = OrderedDict()
                attr['dst_shape'] = '-1,768'
                new_node.attr = attr

                matmul_node.input_tensors[0] = new_node.output_tensors[0]
                all_dest_op.append(matmul_node.name)
                
            # only insert one node to reshape the encoder_hidden_satates.
            new_node.output_tensors[0].dest_op = all_dest_op
            assert first_matmul_node_idx != -1
            model.insert_nodes(first_matmul_node_idx, [new_node])

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The neural engine compile module."""

from collections import OrderedDict
from .loaders.loader import Loader
from .extractors.extractor import Extractor
from .sub_graph.subgraph_matcher import SubGraphMatcher
from .graph_utils import get_model_fwk_name, construct_node, pattern_mapping
from .ops.tensor import Tensor
from .graph import Graph
import numpy as np
from . import graph_utils as util

COMPILES = OrderedDict({
    'loader': Loader,
    'extractor': Extractor,
    'sub_graph': SubGraphMatcher,
})


class autocast:

    def __init__(self, cast_type: str) -> None:
        util.autocast_init()
        self.prev_cast_type = util.get_autocast_info()['cast_type']
        self.cast_type = cast_type

    def __enter__(self) -> None:
        self.prev_cast_type = util.get_autocast_info()['cast_type']
        util.set_autocast("cast_type", self.cast_type)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        util.set_autocast("cast_type", self.prev_cast_type)


def _config_validation(config):
    """The validation of the input config."""
    if config == None:
        return None

    import yaml
    from schema import Schema

    with open(config, 'r') as conf_file:
        conf = yaml.safe_load(conf_file)

    conf_schema = Schema({
        'pattern_switch': Schema({str: bool}, error='You should provide correct fused_patterns.')
    })

    return conf_schema.validate(conf)


def start_pipeline(model, config=None):
    """The compile pipeline."""
    compile_list = []
    # initialize the compile
    for compile_type in COMPILES.keys():
        compile_ = COMPILES[compile_type]()
        compile_list.append(compile_)
    # convert the model
    for compile_ in compile_list:
        model = compile_(model, pattern_config=config)
    return model


def compile(model, config=None) -> Graph:
    """The compile interface.

    Firstly, use model loader to get the computation graph with corresponding framework.
    The graph contains nodes and edges, the node is op and the edge is the tensor.
    Then extract the ops in the graph and pack them to our form.
    Next exploit these above ops to consist sub-graph, which can see as "a new big op", like LayerNorm.

    Note:
        There may have different computation flow in one subgraph.
    Finally, convert them to .yaml file and .bin file for model configuration and inference.
    """
    from .graph import Graph
    try:
        util.get_autocast_info()
    except:
        util.autocast_init()
    if not isinstance(model, Graph):
        if get_model_fwk_name(model) == 'neural engine':
            graph = Graph()
            graph.graph_init(model + '/conf.yaml', model + '/model.bin')
            model = graph
        else:
            config = _config_validation(config)
            model = start_pipeline(model, config=config)
    if util.get_autocast_info()['cast_type'] == "dynamic_int8":
        model = _dynamic_quantization(model)
    return model


def _insert_Q(graph: Graph):

    def quantize_src_tensor(input_tensor_idx: int, quantize_type: str):
        input_tensor = node.input_tensors[input_tensor_idx]
        node_name = node.input_tensors[input_tensor_idx].name + "_quant" + str(input_tensor_idx)
        quant_output = Tensor(name=input_tensor.name + "_quant",
                              source_op=[node_name],
                              dest_op=[node.name],
                              dtype=quantize_type)
        quant_min = Tensor(name=input_tensor.name + "_min",
                           source_op=[node_name],
                           dest_op=[node.name],
                           dtype='fp32')
        quant_scale = Tensor(name=input_tensor.name + "_scale",
                             source_op=[node_name],
                             dest_op=[node.name],
                             dtype='fp32')
        quantize_op = construct_node(node_name=node_name,
                                     op_type='Quantize',
                                     input_tensors=[input_tensor],
                                     output_tensors=[quant_output, quant_min, quant_scale],
                                     attr=OrderedDict({'output_dtype': quantize_type}))
        try:
            quant_id = graph.get_node_id(node_name)
            quant_node = graph.nodes[quant_id]
            for output_tensor_idx in range(3):
                quant_node.output_tensors[output_tensor_idx].dest_op.append(node.name)
        except ValueError:
            graph.insert_nodes(graph.get_node_id(node.name), [quantize_op])
        node.input_tensors[input_tensor_idx] = quant_output
        node.input_tensors.extend([quant_min, quant_scale])

    def quantize_weight_tensor(weight_tensor: Tensor):
        tensor_min = Tensor(name=node.input_tensors[1].name + "_min",
                            source_op=[],
                            dest_op=[node.name],
                            dtype='fp32')
        tensor_scale = Tensor(name=node.input_tensors[1].name + "_scale",
                              source_op=[],
                              dest_op=[node.name],
                              dtype='fp32')
        weight_perm = node.attr['src1_perm']
        weight_data = weight_tensor.data
        if weight_perm == "0,1":  #1,0?
            tensor_min.data = np.min(weight_data, axis=-1).astype(np.float32)
            max_data = np.max(weight_data, axis=-1).astype(np.float32)
            # tensor_min.data = np.minimum(tensor_min.data, 0.)
            # tensor_max.data = np.maximum(tensor_max.data, 0.)
            tensor_scale.data = np.maximum(np.fabs(tensor_min.data), np.fabs(max_data))
            tensor_scale.data = np.where(tensor_scale.data == 0., 0.,
                                         127. / tensor_scale.data)  #only support s8 now
            zero_point = 0.
            weight_tensor.data = \
                (np.round((weight_tensor.data).T * tensor_scale.data).astype(np.int8)).T
        else:
            tensor_min.data = np.min(weight_data, axis=0).astype(np.float32)
            max_data = np.max(weight_data, axis=0).astype(np.float32)
            # tensor_min.data = np.minimum(tensor_min.data, 0.)
            # tensor_max.data = np.maximum(tensor_max.data, 0.)
            tensor_scale.data = np.maximum(np.fabs(tensor_min.data), np.fabs(max_data))
            tensor_scale.data = np.where(tensor_scale.data == 0., 0.,
                                         127. / tensor_scale.data)  #only support s8 now
            zero_point = 0.
            weight_tensor.data = np.round(weight_tensor.data * tensor_scale.data).astype(np.int8)
        tensor_min.shape = list(tensor_min.data.shape)
        tensor_scale.shape = list(tensor_scale.data.shape)
        weight_tensor.dtype = 's8'
        weight_tensor.name = node.input_tensors[1].name + "_quant"
        node.input_tensors.extend([tensor_min, tensor_scale])

    for node in filter(lambda x: x.op_type in ["InnerProduct", "Matmul"], reversed(graph.nodes)):
        reshape_tensor = []
        if "reshape_dims" in node.attr:
            reshape_tensor = [node.input_tensors[-1]]
            node.input_tensors = node.input_tensors[:-1]
        quantize_src_tensor(0, "s8" if node.op_type == "Matmul" else "u8")
        if isinstance(node.input_tensors[1].data, np.ndarray):
            quantize_weight_tensor(node.input_tensors[1])
        else:
            quantize_src_tensor(1, "s8")
        node.input_tensors.extend(reshape_tensor)


def _fuse_quatize(graph: Graph):
    pattern = {
        "patterns": {
            'in': [[(0, "ANY"), (1, 'Quantize')]],
            'out': [[(0, 'ANY')]]
        },
        'search_mode': 'op_type',
        'node_names': {
            0: 0
        },
        'input_tensors': {
            0: [[{
                0: [0]
            }], [[0], 1]]
        },
        'output_tensors': {
            0: [[{
                1: [0]
            }, {
                1: [1]
            }, {
                1: [2]
            }], [[0, 1, 2], 3]]
        },
        'returns': [0, 1]
    }

    for any_op in ["InnerProduct", "Matmul", "Softmax"]:  #,"LayerNorm"
        now_pattern = pattern.copy()
        now_pattern["patterns"]["in"][0][0] = (0, any_op)
        now_pattern["patterns"]["out"][0][0] = (0, any_op)
        graph, new_node_names, ret_old_nodes = pattern_mapping("fuse_quatize", now_pattern, graph)
        if len(new_node_names) != 0:
            for idx in range(len(new_node_names)):
                new_node_idx = graph.get_node_id(new_node_names[idx][0])
                graph.nodes[new_node_idx].input_tensors = ret_old_nodes[idx][0].input_tensors
                graph.nodes[new_node_idx].attr = ret_old_nodes[idx][0].attr
                graph.nodes[new_node_idx].attr[
                    "output_dtype"] = "u8" if any_op == "Softmax" else ret_old_nodes[idx][1].attr[
                        "output_dtype"]


def _remove_unused_input(graph):
    new_input_tensors = []
    for input_ternsor in graph.nodes[0].output_tensors:
        if input_ternsor.location == [] or input_ternsor.location is None:
            new_input_tensors.append(input_ternsor)
    graph.nodes[0].output_tensors = new_input_tensors


def _dynamic_quantization(fp32_model):
    """
    fp32_model is a engine ir path or a fp32 graph
    """
    # load fp32model
    if isinstance(fp32_model, Graph):
        graph = fp32_model
    else:
        graph = Graph()
        graph.graph_init(fp32_model + '/conf.yaml', fp32_model + '/model.bin')

    _insert_Q(graph)
    _fuse_quatize(graph)
    _remove_unused_input(graph)
    return graph



"""Common classes for different NAS approaches.

NAS class for creating object of different NAS approaches.
NASBase class defines the common methods of different NAS approaches.
"""

#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import os
import shutil

from collections.abc import Iterable
from intel_extension_for_transformers.optimization.config import NASConfig
from neural_compressor.conf.config import Conf
from neural_compressor.experimental.nas.nas_utils import find_pareto_front, NASMethods
from neural_compressor.experimental.nas.search_algorithms import \
    BayesianOptimizationSearcher, GridSearcher, RandomSearcher
from neural_compressor.utils.utility import logger, LazyImport

torch = LazyImport('torch')


class NAS(object):
    """Create object of different NAS approaches.

    Args:
        conf_fname_or_obj (string or obj):
            The path to the YAML configuration file or the object of NASConfig.

    Returns:
        An object of specified NAS approach.
    """

    def __new__(self, conf_fname_or_obj, *args, **kwargs):
        """Create an object of specified NAS approach."""
        if isinstance(conf_fname_or_obj, str) and os.path.isfile(conf_fname_or_obj):
            self.config = Conf(conf_fname_or_obj).usr_cfg
        elif isinstance(conf_fname_or_obj, NASConfig):
            self.config = conf_fname_or_obj.config
        else:  # pragma: no cover
            raise NotImplementedError(
                "Please provide a str path to the config file."
            )
        assert self.config.nas is not None, "nas section must be set"
        if isinstance(self.config.nas.approach, str) and \
                self.config.nas.approach.lower() in NASMethods:
            method = self.config.nas.approach.lower()
        else:  # pragma: no cover
            logger.warning(
                "NAS approach not set in config, use default NAS approach, i.e. Basic."
            )
            method = 'basic'
        return NASMethods[method](conf_fname_or_obj, *args, **kwargs)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry
from .tensor import Tensor
from ..graph_utils import list2str


# tf.strided_slice(input_, begin, end, strides=None, begin_mask=0, end_mask=0,
# ellipsis_mask=0,new_axis_mask=0, shrink_axis_mask=0, var=None, name=None)
@operator_registry(operator_type='StridedSlice')
class StridedSlice(Operator):
    """Parse the StridedSlice operator to the neural engine."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        """Extract the node attr from tensorflow."""
        if framework == 'tensorflow':
            self._attr['begin_mask'] = node.attr['begin_mask'].i
            self._attr['ellipsis_mask'] = node.attr['ellipsis_mask'].i
            self._attr['end_mask'] = node.attr['end_mask'].i
            self._attr['new_axis_mask'] = node.attr['new_axis_mask'].i
            self._attr['shrink_axis_mask'] = node.attr['shrink_axis_mask'].i

            begin = None
            end = None
            strides = None
            if self.input_tensors[1].source_op == []:
                begin = self.input_tensors[1]
                self._attr['begin'] = list2str(begin.data)
            if self.input_tensors[2].source_op == []:
                end = self.input_tensors[2]
                self._attr['end'] = list2str(end.data)
            if self.input_tensors[3].source_op == []:
                strides = self.input_tensors[3]
                self._attr['strides'] = list2str(strides.data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Distillation: set criterion mode to distillation."""
from enum import Enum
from typing import List


class Criterion(object):
    """Criterion class for distillation."""
    def __init__(
        self,
        name: str = "KNOWLEDGELOSS",
        temperature: float = 1.0,
        loss_types: List = ['CE', 'CE'],
        loss_weight_ratio: List = [0.5, 0.5],
        layer_mappings: List = None,
        add_origin_loss: bool = False
    ):
        """Init a Criterion object."""
        self.name = name
        self.temperature = temperature
        self.loss_types = loss_types
        self.loss_weight_ratio = loss_weight_ratio
        self.layer_mappings = layer_mappings
        self.add_origin_loss = add_origin_loss


class DistillationCriterionMode(Enum):
    """Criterion mode class for distillation."""
    KNOWLEDGELOSS = "KnowledgeDistillationLoss"
    INTERMEDIATELAYERSLOSS = "IntermediateLayersKnowledgeDistillationLoss"
    


SUPPORTED_DISTILLATION_CRITERION_MODE = \
    set([approach.name for approach in DistillationCriterionMode])



import os
import unittest
import shutil
import neural_compressor.adaptor.pytorch as nc_torch
from intel_extension_for_transformers.optimization.pipeline import pipeline
from packaging.version import Version
from transformers import (
    set_seed,
    AutoConfig,
    AutoTokenizer,
)

os.environ["WANDB_DISABLED"] = "true"
os.environ["GLOG_minloglevel"] = "2"
os.environ["DISABLE_MLFLOW_INTEGRATION"] = "true"

PT_VERSION = nc_torch.get_torch_version()
MODEL_NAME = "distilbert-base-uncased-finetuned-sst-2-english"

message = "The output scores should be close to 0.9999."


class TestPipeline(unittest.TestCase):
    @classmethod
    def tearDownClass(cls):
        shutil.rmtree('./tmp_trainer', ignore_errors=True)

    def test_fp32_pt_model(self):
        text_classifier = pipeline(
            task="text-classification",
            model="distilbert-base-uncased-finetuned-sst-2-english",
            framework="pt",
        )
        outputs = text_classifier("This is great !")
        self.assertAlmostEqual(outputs[0]['score'], 0.9999, None, message, 0.0001)

    def test_int8_pt_model(self):
        text_classifier = pipeline(
            task="text-classification",
            model="Intel/distilbert-base-uncased-finetuned-sst-2-english-int8-static",
            framework="pt",
        )
        outputs = text_classifier("This is great !")
        self.assertAlmostEqual(outputs[0]['score'], 0.8, None, message, 0.3)


@unittest.skipIf(PT_VERSION.release >= Version("1.12.0").release,
    "Please use PyTroch 1.11 or lower version for executor backend")
class TestExecutorPipeline(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        set_seed(42)
        from test_quantization import TestQuantization
        TestQuantization.setUpClass()
        cls.config = AutoConfig.from_pretrained(MODEL_NAME)
        cls.tokenizer = AutoTokenizer.from_pretrained(MODEL_NAME)

    @classmethod
    def tearDownClass(cls):
        from test_benchmark import TestExecutorBenchmark
        TestExecutorBenchmark.tearDownClass()

    def test_fp32_executor_model(self):
        text_classifier = pipeline(
            task="text-classification",
            config=self.config,
            tokenizer=self.tokenizer,
            model='tmp_trainer/fp32-model.onnx',
            model_kwargs={'backend': "executor"},
        )
        outputs = text_classifier(
            "But believe it or not , it 's one of the most "
            "beautiful , evocative works I 've seen ."
        )
        self.assertAlmostEqual(outputs[0]['score'], 0.9999, None, message, 0.0001)

    def test_int8_executor_model(self):
        text_classifier = pipeline(
            task="text-classification",
            config=self.config,
            tokenizer=self.tokenizer,
            model='tmp_trainer/int8-model.onnx',
            model_kwargs={'backend': "executor"},
        )
        outputs = text_classifier(
            "But believe it or not , it 's one of the most "
            "beautiful , evocative works I 've seen ."
        )
        # increase score range to cater data fluctuations.
        self.assertAlmostEqual(outputs[0]['score'], 0.8, None, message, 0.3)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine onnx utils."""

import sys
import numpy as np
import re
from collections import namedtuple, OrderedDict
from . import logger
from .ops.tensor import Tensor
from . import graph_utils as util

def get_children(node, input_name_to_nodes=None):
    """Get the node's output nodes in the graph."""
    if input_name_to_nodes is None:
        input_name_to_nodes = {}

    children = []
    for output in node.output:
        if output in input_name_to_nodes:
            for child in input_name_to_nodes[output]:
                children.append(child)
    return children

def get_node_children_names(model, node):
    """Get the node's output nodes' name in the graph.

    Args:
        model: ONNXModel
        node: NodeProto in onnx model

    Returns:
        outputs: names list
    """
    output_nodes = get_children(node)
    outputs = [node.name for node in output_nodes]
    return outputs


def get_initializer_children_names(model, initializer):
    """Get the initializer's output nodes' name in the graph.

    Args:
        model: ONNXModel
        initializer: initializer in onnx model

    Returns:
        outputs: names list
    """
    output_nodes = []
    for node in model.graph.node:
        for node_input in node.input:
            if node_input == initializer.name:
                output_nodes.append(node)
    outputs = [node.name for node in output_nodes]
    return outputs


def graph_node_names_details(model):
    """Parse the graph nodes ans get the graph_nodes_dict.

    Be used for Grpah class with cerating a new graph.
    The node_name is the key, node in value is for getting the Const
    tensor value and the input_tensor source op; output_names in value
    is the node ouput name list; outputs in value is for output_tensor dest op
    Args:
        model: ONNXModel
        
    Returns:
        node_names_details: the graph node info dict

    """
    node_details = namedtuple('node_details', ['node', 'outputs'])
    node_names_details = {}
    for initializer in model.graph.initializer:
        initializer_name = initializer.name
        each_node = node_details(node=initializer, outputs=[])
        if initializer_name not in node_names_details:
            each_node.outputs.extend(get_initializer_children_names(model, initializer))
            node_names_details[initializer_name] = each_node
    for node in model.graph.node:
        node_name = node.name
        output_names = node.output
        for index, output_name in enumerate(output_names):
            output_names[index] = util.names_from_input(output_name)[1]
        # onnx output has different name from node name
        for output_name in output_names:
            if output_name not in node_names_details:
                node_names_details[output_name] = node_name
        each_node = node_details(node=node, outputs=[])
        if node_name not in node_names_details:
            each_node.outputs.extend(get_node_children_names(model, node))
            node_names_details[node_name] = each_node
    for graph_input in model.graph.input:
        outputs = []
        node_name = graph_input.name
        for k, v in node_names_details.items():
            try:
                if node_name in v.node.input:
                    outputs.append(k)
            except BaseException:
                continue
        each_node = node_details(node=graph_input, outputs=outputs)
        # if node_name not in node_names_details:
        node_names_details[node_name] = each_node

    return node_names_details

def is_supported_onnx_graph(graph):
    """Check if the onnx graph supported."""
    for node in graph.node:
        if not is_supported_onnx_node(node.op_type):
            return False
    return True

def is_supported_onnx_node(node_name):
    """Check if the node type is supported."""
    supported_ops_type = ["Add", "Softmax", "Slice", "ReduceMean", "Reshape",
                        "Concat", "Gather", "QuantizeLinear", "Transpose", "MatMul",
                        "Sqrt", "Unsqueeze", "Shape", "Erf", "Pow", "DequantizeLinear",
                        "Cast", "Tanh", "Div", "Mul", "Sub", "Constant", "Relu", "Conv",
                        "Identity", "Split", "TopK"]
    if node_name in supported_ops_type:
        return True
    else:
        return False

def change_num_name(tensor_name):
    """For number string."""
    try:
        if str(int(tensor_name)) == tensor_name:
            tensor_name += '_tensor'
    except BaseException:
        pass
    return tensor_name


def bias_to_int32(bias_node, a_scale, b_scale):
    """Convert the int8 bias to int32 bias.

    Args:
        bias_node: bias_add in graph (from onnx framework)
        a_scale: matmul node input matrice a scale tensor
        b_scale: matmul node input matrice b scale tensor
        model: Grpah class

    Returns:
        int32 bias numpy array

    fp32_bias = (int8_bias - int8_bias_zero_point) * int8_bias_scale
    int32_bias = fp32_bias / (a_scale * b_scale)
    """
    dtype_list = [np.int8, np.uint8]
    input_tensors = bias_node.input_tensors
    if input_tensors[0].source_op == []:
        int8_bias = input_tensors[0].data
        int8_bias_scale = input_tensors[1].data
        int8_bias_zero_point = input_tensors[2].data
    else:
        int8_bias = input_tensors[3].data
        int8_bias_scale = input_tensors[4].data
        int8_bias_zero_point = input_tensors[5].data

    if int8_bias.dtype not in dtype_list:
        return
    int8_bias = int8_bias.astype(np.int32)
    fp32_bias = ((int8_bias - int8_bias_zero_point) * int8_bias_scale).astype(np.float32)
    a_scale = float(a_scale)
    b_scale = float(b_scale)
    int32_bias = np.round((fp32_bias / (a_scale * b_scale))).astype(np.int32)

    return int32_bias

def _bf16_tensor_to_array(tensor, base_dir: str = "") -> np.ndarray:
    """Converts a tensor def object to a numpy array.

    Inputs:
        tensor: a TensorProto object.
        base_dir: if external tensor exists, base_dir can help to find the path to it
    Returns:
        arr: the converted array.
    """
    onnx = util.LazyImport('onnx')
    from onnx import TensorProto
    from onnx.external_data_helper import load_external_data_for_tensor, uses_external_data
    from onnx.numpy_helper import convert_endian

    if tensor.HasField("segment"):
        raise ValueError("Currently not supporting loading segments.")
    if tensor.data_type == TensorProto.UNDEFINED:
        raise TypeError("The element type in the input tensor is not defined.")

    tensor_dtype = tensor.data_type
    np_dtype = np.dtype('float16')
    storage_np_dtype = np.dtype('uint16')
    storage_field = 'int32_data'
    dims = tensor.dims

    # Load raw data from external tensor if it exists
    if uses_external_data(tensor):
        load_external_data_for_tensor(tensor, base_dir)

    if tensor.HasField("raw_data"):
        # Raw_bytes support: using frombuffer.
        if sys.byteorder == 'big':
            # Convert endian from little to big
            convert_endian(tensor)

        return np.frombuffer(
            tensor.raw_data,
            dtype=np_dtype).reshape(dims)
    else:
        # float16/bfloat16 is stored as int32 (uint16 type); Need view to get the original value
        return (
            np.asarray(
                tensor.int32_data,
                dtype=np.uint16)
            .reshape(dims)
            .view(np.float16))


def onnx_extract_operator(node, framework_model, nodes_dict, engine_graph=None):
    """Decorate the operator in onnx.

    Args:
        node: NodeProto
        framework_model: ONNXModel
        nodes_dict: dict, return value from graph_node_names_details
        tf_dtypes: dict, for get the dtype string
        engine_graph: Engine Graph class

    Returns:
        op_type: node op type
        input_tensors: Tensor list, contains the node input tensors info
        output_tensors: Tensor list, contains the node output tensor info
    """
    onnx = util.LazyImport('onnx')
    from onnx import TensorProto
    from onnx.numpy_helper import to_array
    op_type = node.op_type
    input_tensors = []
    output_tensors = []

    """Input_tensors

    Note:
        each input_tensor has its own soure op, but same dest op
        so both have single string
    """
    input_names = []
    input_tensor_names = node.input
    for input_tensor_name in input_tensor_names:
        # The resize op could have the third input name but no second input.
        if input_tensor_name == '':
            continue
        origin_tensor_name, input_tensor_name = util.names_from_input(input_tensor_name)
        try:
            pre_node = nodes_dict[nodes_dict[input_tensor_name]].node
        except BaseException:
            # if origin_tensor_name in nodes_dict:
            pre_node = nodes_dict[origin_tensor_name].node
        
        data = None
        if pre_node in framework_model.graph.initializer:
            if pre_node.data_type == TensorProto.BFLOAT16:
                data = _bf16_tensor_to_array(pre_node)
            else:
                data = to_array(pre_node)
        else:
            if (pre_node not in framework_model.graph.input) and (pre_node.op_type == 'Constant'):
                data = to_array(pre_node.attribute[0].t)
        if isinstance(data, np.ndarray):
            if data.dtype == np.int64:
                data = data.astype(np.int32)
            dtype = util.get_data_dtype(data)
            shape = list(data.shape) if data.shape != () else [1]
            input_tensor = Tensor(name=input_tensor_name,
                                  source_op=[],
                                  dest_op=[node.name],
                                  shape=shape,
                                  data=data,
                                  dtype=dtype
                                  )
            input_tensors.append(input_tensor)

        else:
            input_tensor = Tensor(name=input_tensor_name,
                                  source_op=[pre_node.name],
                                  dest_op=[node.name],
                                  shape=None,
                                  data=None,
                                  dtype=None
                                  )
            input_tensors.append(input_tensor)
        input_names.append(node.name)

    """Output_tensors.
    
    Note:
        in onnx, NodeProto has the output attribute
    """
    output_tensor_names = node.output
    for output_tensor_name in output_tensor_names:
        output_tensor_name = util.names_from_input(output_tensor_name)[1]
        output_tensor = Tensor(name=output_tensor_name,
                               source_op=[node.name],
                               dest_op=nodes_dict[node.name].outputs,
                               shape=None,
                               data=None,
                               dtype=None
                               )

        output_tensors.append(output_tensor)

    return op_type, input_tensors, output_tensors


ONNX_DTYPE_ID = {1: 'fp32',
                7: 'int32',
                9: 'bool',
                6: 'int32',}

# key is the optimize level (higher means you can use more passes)
# value is the pass name
# for more passes, see: https://github.com/onnx/optimizer/blob/master/onnxoptimizer/pass_registry.h
ONNX_OPTIMIZER_PASS = {1: 'eliminate_identity',
                       2: 'eliminate_shape_gather',
                      }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The PositionEmbeddings pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util


@pattern_registry(pattern_type='PositionEmbeddings')
class PositionEmbeddings(Pattern):
    """The PositionEmbeddings pattern.

    Fuse the original sub-graph into the custom acceleration 'PositionEmbeddings' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'PositionEmbeddings': [
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'StridedSlice'), (2, 'Pack'), (3, 'Slice'),
                                (5, 'Reshape')], [(1, 'StridedSlice'), (4, 'Pack'),
                                                  (5, 'Reshape')]],
                        'out': [[(0, 'Reshape'), (1, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 'position_embeddings/after/reshape',
                        1: 5
                    },
                    'input_tensors': {
                        0: [[{
                            3: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]],
                        1: [[], [[], 1]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[{
                            5: [0]
                        }], [[0], 1]]
                    },
                    'returns': [3]
                },

                # distil_bert_base
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Cast'), (3, 'Range'),
                                (4, 'Unsqueeze'), (6, 'Expand'), (7, 'Gather')],
                                [(), (5, 'Shape'), (6, 'Expand')]],
                        'out': [[(0, 'Reshape'), (1, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 'position_embeddings/after/reshape',
                        1: 7
                    },
                    'input_tensors': {
                        0: [[{
                            7: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]],
                        1: [[], [[], 1]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[{
                            7: [0]
                        }], [[0], 1]]
                    },
                    'returns': [7]
                },

                # geminet
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'),
                                (9, 'ConstantOfShape'), (10, 'NonZero'), (11, 'Transpose'),
                                (12, 'Squeeze'), (13, 'Unsqueeze'), (14, 'Expand'),
                                (15, 'Gather')],
                                [(0, 'Shape'), (3, 'Gather'), (4, 'Unsqueeze'), (5, 'Concat'),
                                (6, 'Reshape'), (7, 'Equal'), (8, 'Where'), (14, 'Expand')]],
                        'out': [[(0, 'Reshape'), (1, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 'position_embeddings/after/reshape',
                        1: 15
                    },
                    'input_tensors': {
                        0: [[{
                            15: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]],
                        1: [[], [[], 1]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[{
                            15: [0]
                        }], [[0], 1]]
                    },
                    'returns': [15]
                },

                # bert_base_sparse
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Add'), (3, 'Unsqueeze'),
                                (4, 'Slice'), (5, 'Gather')]],
                        'out': [[(0, 'Reshape'), (1, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 'position_embeddings/after/reshape',
                        1: 5
                    },
                    'input_tensors': {
                        0: [[{
                            5: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]],
                        1: [[], [[], 1]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[{
                            5: [0]
                        }], [[0], 1]]
                    },
                    'returns': [5]
                },

                # bert_mini_int8
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'),
                                (3, 'Slice'), (4, 'Gather')]],
                        'out': [[(0, 'Reshape'), (1, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 'position_embeddings/after/reshape',
                        1: 4
                    },
                    'input_tensors': {
                        0: [[{
                            4: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]],
                        1: [[], [[], 1]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[{
                            4: [0]
                        }], [[0], 1]]
                    },
                    'returns': [4]
                },

                # opennmt encoder
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (3, 'Slice'),
                                (4, 'Add')]],
                        'out': [[(0, 'Reshape'), (1, 'Add')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 3,
                        1: 4,
                    },
                    'input_tensors': {
                        0: [[{
                            3: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]],
                        1: [[{
                            4: [0]
                        }], [[0], 2]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[{
                            4: [0]
                        }], [[0], 1]]
                    },
                    'returns': [3]
                },

                # bert_base_mrpc
                {
                    'patterns': {
                        'in': [[(0, 'Slice'), (1, 'Reshape')]],
                        'out': [[(0, 'Reshape'), (1, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 'position_embeddings/after/reshape',
                        1: 1
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]],
                        1: [[], [[], 1]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[{
                            1: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0]
                },
            ]
        }

        def _set_attr(hidden_size, node_names, model, batch_idx=0):
            attr1 = OrderedDict()
            if batch_idx == 0:
                attr1['dst_shape'] = '1,-1,' + str(hidden_size)
                attr1['dims'] = 1
            else:
                attr1['dst_shape'] = '-1,1,' + str(hidden_size)
                attr1['dims'] = 0
            if batch_idx == 0:
                attr2 = OrderedDict()
                attr2['dst_shape'] = '1,-1'

            reshape_0_node_idx = model.get_node_id(node_names[0])
            model.nodes[reshape_0_node_idx].attr = attr1

            if batch_idx == 0:
                reshape_1_node_idx = model.get_node_id(node_names[1])
                model.nodes[reshape_1_node_idx].attr = attr2

        def _remove_assert(pattern, model):
            rm_rets = util.search_pattern(pattern, model)
            for ret in rm_rets:
                model.remove_nodes(ret[:-1])
            return model

        for i in range(0, len(pattern_mapping_config['PositionEmbeddings'])-1):
            pattern_dict = pattern_mapping_config['PositionEmbeddings'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("PositionEmbeddings", 
                                                                        pattern_dict, model)
            if len(new_node_names) != 0:
                for j in range(len(new_node_names)):
                    slice_node = ret_old_nodes[j][0]
                    hidden_size = int(slice_node.input_tensors[0].shape[-1])
                    batch_idx = 0
                    if model.get_node_by_name(new_node_names[j][0]).op_type == "Reshape" and \
                        model.get_node_by_name(new_node_names[j][1]).op_type != "Reshape":
                        batch_idx = 1
                    _set_attr(hidden_size, new_node_names[j], model, batch_idx=batch_idx)

                return model

        # bert_base_mrpc
        pattern_dict = pattern_mapping_config['PositionEmbeddings'][-1]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("PositionEmbeddings", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            for i in range(len(new_node_names)):
                slice_node = ret_old_nodes[i][0]
                hidden_size = int(slice_node.input_tensors[0].shape[-1])
                _set_attr(hidden_size, new_node_names[i], model)
            p = [[(0, 'LessEqual'), (1, 'All'), (2, 'Assert')]]
            model = _remove_assert(p, model)

            return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The TextEncoder_WordEmbedding pattern."""

from .pattern import Pattern, pattern_registry
from collections import OrderedDict
from .. import graph_utils as util
from .. import logger


@pattern_registry(pattern_type='TextEncoder_WordEmbedding')
class TextEncoder_WordEmbedding(Pattern):
    """The TextEncoder_WordEmbedding pattern.

    Fuse the original sub-graph into the custom acceleration 'TextEncoder_WordEmbedding' graph.
    The search strategy is based on the following pattern mapping configs for the stable textEncoderV1-5.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'TextEncoder_WordEmbedding': [
                {
                    'patterns': {
                        'in': [[(0, 'Unsqueeze'), (1, 'Concat'), (2, 'Reshape'), (3, 'Gather'), (4, 'Add')]],
                        'out': [[(0, 'Reshape'), (1, 'Reshape'), (2, 'Gather'), (3, 'Reshape'),
                                 (4, 'Reshape'), (5, 'Add')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 2,
                        1: 'textEncoder_word_embeddings/reshape',
                        2: 3,
                        3: 'textEncoder_word_embeddings/after/reshape',
                        4: 'textEncoder_word_embeddings/add_reshape',
                        5: 4
                    },
                    'input_tensors': {
                        0: [[{
                            2: [0]
                        }], [[0], 1]],
                        1: [[], [[], 1]],
                        2: [[{
                            3: [0]
                        }], [[1], 2]],
                        3: [[{
                            2: [0]
                        }], [[1], 2]],
                        4: [[{
                            2: [0]
                        }], [[1], 2]],
                        5: [[{
                            4: [1]
                        }], [[1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            2: [0]
                        }], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[], [[], 1]],
                        3: [[], [[], 1]],
                        4: [[], [[], 1]],
                        5: [[{
                            4: [0]
                        }], [[0], 1]],
                    },
                    'returns': [3, 2]
                },
            ]
        }

        def _set_attr(hidden_size, axis, batch_dims, node_names, model):
            attr0 = OrderedDict()
            attr0['dst_shape'] = '1,-1'
            attr1 = OrderedDict()
            attr1['dst_shape'] = -1
            attr2 = OrderedDict()
            attr2['axis'] = axis
            attr2['batch_dims'] = batch_dims
            attr3 = OrderedDict()
            attr3['dst_shape'] = '-1,-1,' + str(hidden_size)
            attr3['dims'] = '0,1'
            attr4 = OrderedDict()
            attr4['dst_shape'] = '-1,-1,' + str(hidden_size)
            attr4['dims'] = '0,1'
            attr4['mul'] = '1,2'

            reshape_node_idx = model.get_node_id(node_names[0])
            model.nodes[reshape_node_idx].attr = attr0

            reshape_0_node_idx = model.get_node_id(node_names[1])
            model.nodes[reshape_0_node_idx].attr = attr1

            gather_node_idx = model.get_node_id(node_names[2])
            model.nodes[gather_node_idx].attr = attr2

            reshape_1_node_idx = model.get_node_id(node_names[3])
            model.nodes[reshape_1_node_idx].attr = attr3

            reshape_2_node_idx = model.get_node_id(node_names[4])
            model.nodes[reshape_2_node_idx].attr = attr4

        for i in range(len(pattern_mapping_config['TextEncoder_WordEmbedding'])):
            pattern_dict = pattern_mapping_config['TextEncoder_WordEmbedding'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("TextEncoder_WordEmbedding",
                                                                        pattern_dict, model)

            if len(new_node_names) != 0:
                logger.info('TextEncoder_WordEmbedding mathched...')
                logger.debug('TextEncoder_WordEmbedding = {}'.format(new_node_names))
                for j in range(len(new_node_names)):
                    gatherv2_node = ret_old_nodes[j][0]
                    hidden_size = int(gatherv2_node.input_tensors[0].shape[-1])
                    axis = gatherv2_node.attr['axis']
                    batch_dims = gatherv2_node.attr['batch_dims']
                    _set_attr(hidden_size, axis, batch_dims, new_node_names[j], model)

                return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The GenerateSequence Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util


@pattern_registry(pattern_type='GenerateSequence')
class GenerateSequence(Pattern):
    """The GenerateSequence pattern.

    Fuse the original sub-graph into the custom acceleration 'GenerateSequence' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """

    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'GenerateSequence': [{
                'patterns': {
                    'in': [[(0, "Shape"), (1, 'Gather'), (2, "Unsqueeze"), (3, "Concat"),
                            (4, "Shape"), (5, "ConstantOfShape"), (6, "Expand"), (7, "Tile")],
                           [(), (8, 'Shape'), (9, 'Gather'), (10, 'Cast'), (11, 'Range'),
                            (12, "Unsqueeze"), (6, "Expand")],
                           [(1, 'Gather'), (13, "Unsqueeze"), (14, "Concat"), (7, "Tile")]],
                    'out': [[(0, 'LatRange')]]
                },
                'search_mode': 'op_type',
                'node_names': {
                    0: 11
                },
                'input_tensors': {
                    0: [[{
                        'input_data': [0]
                    }], [[0], 1]],
                },
                'output_tensors': {
                    0: [[{
                        7: [0]
                    }], [[0], 1]],
                },
                'returns': [11]
            },
            {
                'patterns': {
                    'in': [[(0, "Shape"), (1, 'Gather'), (2, "Unsqueeze"), (3, "Concat"),
                            (7, "Tile")],
                           [(0, "Shape"), (4, 'Gather'), (5, 'Range'),
                            (6, "Unsqueeze"), (7, "Tile")]],
                    'out': [[(0, 'LatRange')]]
                },
                'search_mode': 'op_type',
                'node_names': {
                    0: 5
                },
                'input_tensors': {
                    0: [[{
                        'input_data': [0]
                    }], [[0], 1]],
                },
                'output_tensors': {
                    0: [[{
                        7: [0]
                    }], [[0], 1]],
                },
                'returns': [5, 0]
            }                   
            ]
        }
        collect_node = []

        for i in range(len(pattern_mapping_config['GenerateSequence'])):
            pattern_dict = pattern_mapping_config['GenerateSequence'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping(
                "GenerateSequence", pattern_dict, model)
            if len(new_node_names) != 0:
                for j in range(len(new_node_names)):
                    old_node = ret_old_nodes[j][0]
                    attr = OrderedDict()
                    attr["start"] = int(old_node.input_tensors[0].data)
                    attr["step"] = int(old_node.input_tensors[2].data)
                    new_node_idx = model.get_node_id(new_node_names[j][0])
                    model.nodes[new_node_idx].attr = attr
                    
                    if i == 1:
                        collect_node.append(ret_old_nodes[j][1])
                model.insert_nodes(10, collect_node)
                return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The TextEncoder_QReshape Pattern."""

from .pattern import Pattern, pattern_registry
from collections import OrderedDict
from .. import graph_utils as util
from .. import logger


@pattern_registry(pattern_type='TextEncoder_QReshape')
class TextEncoder_QReshape(Pattern):
    """The TextEncoder_QReshape pattern.

    Fuse the original sub-graph into the custom acceleration 'TextEncoder_QReshape' graph.
    The search strategy is based on the following pattern mapping configs for the stable diffusionV1-5.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'TextEncoder_QReshape': [
                # for text encoder q_proj reshape(Reshape_2 & Reshape_8)
                {
                    'patterns': {
                        'in': [[(0, 'Unsqueeze'), (2, 'Concat'), (3, 'Reshape'), (4, 'Transpose')],
                               [(), (1, 'Unsqueeze'), (2, 'Concat')]],
                        'out': [[(0, 'Reshape'), (1, 'Transpose')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 3,
                        1: 4
                    },
                    'input_tensors': {
                        0: [[{
                            3: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]],
                        1: [[], [[], 1]],
                    },
                    'output_tensors': {
                        0: [[{
                            3: [0]
                        }], [[0], 1]],
                        1: [[{
                            4: [0]
                        }], [[0], 1]],
                    },
                    'returns': [2, 4]
                },
            ]
        }

        for i in range(len(pattern_mapping_config['TextEncoder_QReshape'])):
            pattern_dict = pattern_mapping_config['TextEncoder_QReshape'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("TextEncoder_QReshape", pattern_dict,
                                                                        model)

            if len(new_node_names) != 0:
                logger.info('TextEncoder_QReshape mathched...')
                logger.debug('TextEncoder_QReshape = {}'.format(new_node_names))
                for j in range(len(new_node_names)):

                    pack_node = ret_old_nodes[j][0]
                    head_size = int(pack_node.input_tensors[-1].data)
                    head_size_idx = 3


                    head_num_idx = False
                    if pack_node.input_tensors[1].data != None:
                        head_num_idx = 1
                        head_num = int(pack_node.input_tensors[1].data)
                    elif pack_node.input_tensors[2].data != None:
                        head_num_idx = 2
                        head_num = int(pack_node.input_tensors[2].data)

                    assert head_num_idx != False

                    attr = OrderedDict()
                    attr['dst_shape'] = []
                    for i in range(len(pack_node.input_tensors)):
                        if i == head_num_idx:
                            attr['dst_shape'].append(str(head_num))
                        elif i == head_size_idx:
                            attr['dst_shape'].append(str(head_size))
                        else:
                            attr['dst_shape'].append('-1')

                    attr['dst_shape'] = ','.join(attr['dst_shape'])
                    attr['dims'] = '0,1'
                    
                    reshape_node_idx = model.get_node_id(new_node_names[j][0])
                    model.nodes[reshape_node_idx].attr = attr

                    assert ret_old_nodes[j][1].op_type == 'Transpose'
                    mat_node_idx = model.get_node_id(new_node_names[j][1])
                    model.nodes[mat_node_idx].attr = ret_old_nodes[j][1].attr

                return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry
from .tensor import Tensor


# The inputs must be two-dimensional matrices
@operator_registry(operator_type='Gemm')
class Gemm(Operator):
    """Parse the Gemm operator to the neural engine."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        """Extract the node attr from frameworks."""
        self._op_type = 'MatMulWithBias'
        if framework == 'onnxruntime':
            transpose_a = False
            transpose_b = False
            alpha = 1.0
            beta = 1.0
            for attr in node.attribute:
                if attr.name == 'transA':
                    transpose_a = bool(attr.i)
                if attr.name == 'transB':
                    transpose_b = bool(attr.i)
                if attr.name == 'alpha':
                    alpha = attr.f
                if attr.name == 'beta':
                    beta = attr.f
            if transpose_a:
                self._attr['src0_perm'] = '1,0'
            # see OneDNN InnerProduct related requirements
            if not transpose_b:
                self._attr['src1_perm'] = '1,0'
            else:
                self._attr['src1_perm'] = '0,1'
            if alpha != 1.0:
                self._attr['alpha'] = alpha
            if beta != 1.0:
                self._attr['beta'] = beta



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The TorchPaddingSequence pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
import copy
from .. import graph_utils as util
from ..ops.tensor import Tensor


@pattern_registry(pattern_type='TorchPaddingSequence')
class TorchPaddingSequence(Pattern):
    """The TorchPaddingSequence pattern.

    Fuse the original sub-graph into the custom acceleration 'TorchPaddingSequence' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'TorchPaddingSequence': [
                {
                    'patterns': {
                        'in': [[(0, 'View'), (1, 'Slice'), (2, 'Unsqueeze'), (3, 'Unsqueeze'),
                                (4, 'Slice'), (5, 'Rsub'), (6, 'Mul')]
                               ],
                        'out': [[(0, 'PaddingSequence')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 6
                    },
                    'input_tensors': {
                        0: [[{
                             0: [0]
                        }], [[0], 1]]
                    },
                    'output_tensors': {
                        0: [[{
                            6: [0]
                        }], [[0], 1]]
                    },
                    'returns': []
                }

            ]
        }

        def _make_padding_sequence_node(tensor_idx, hidden_size, model):
            # Models with different size may have different nums of self-attention head.
            # In Google Bert, the nums_head = hidden_size // 64.
            # But in some other models, like minilm, they has smaller attention head channel.
            # Use this dict to maintain the attr of padding_sequence operation.
            # The last key is for debug and tell the info of unknown model.
            heads_num_dict = {4096: 16, 1024: 16, 768: 12, 384: 12, 256: 4, 128: 2, -1: -1}
            node_name = 'padding_sequence'
            heads_num = int(heads_num_dict[hidden_size])
            input_data = model.nodes[0]
            if len(input_data.output_tensors) > tensor_idx:
                input_tensors = [copy.deepcopy(input_data.output_tensors[tensor_idx])]
            else:
                input_tensors = [Tensor()]
            output_tensors = [Tensor(name=node_name + ':0', source_op=[node_name], dest_op=[])]
            attr = OrderedDict()
            attr['dst_shape'] = '-1,' + str(heads_num) + ',0,-1'
            attr['dims'] = 1
            padding_sequence_node = util.construct_node(node_name,
                                                        'PaddingSequence',
                                                        input_tensors=input_tensors,
                                                        output_tensors=output_tensors,
                                                        attr=attr)

            model.insert_nodes(1, [padding_sequence_node])

            return model

        def get_hidden_size(model, p=None, mat_idx=0):
            if p == None:
                p = [[(0, 'InnerProduct'), (1, 'Pow')]]
            match_result = util.search_pattern(p, model)
            if len(match_result) != 0:
                mat_node = model.get_node_by_name(match_result[0][mat_idx])
                # import pdb;pdb.set_trace()
                hidden_size = 4096
                #hidden_size = int(mat_node.input_tensors[1].shape[-1])
                
                #hidden_size = 768
            else:
                hidden_size = -1
            return hidden_size

        if model.framework_modeling_config['framework'] == 'torch':
            pattern_dict = pattern_mapping_config['TorchPaddingSequence'][0]
            hidden_size = get_hidden_size(model)
            model = _make_padding_sequence_node(1, hidden_size, model)
            model, new_node_names, ret_old_nodes = util.pattern_mapping("TorchPaddingSequence",
                                                                        pattern_dict, model)

            if len(new_node_names) != 0:
                assert hidden_size!=-1, "Wrong hidden size in padding_sequence!"
                ps_attr = model.get_node_by_name('padding_sequence').attr
                for j in range(len(new_node_names)):
                    ps_node_id = model.get_node_id(new_node_names[j][0])
                    model.nodes[ps_node_id].attr = ps_attr
                # remove fake node
                model.remove_nodes(['padding_sequence'])
                return model
            else:
                model.remove_nodes(['padding_sequence'])


        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The LayerNormWithTranspose Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
import copy


@pattern_registry(pattern_type='LayerNormWithTranspose')
class LayerNormWithTranspose(Pattern):
    """The LayerNormWithTranspose pattern.

    Fuse the original sub-graph into the custom acceleration 'LayerNormWithTranspose' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'LayerNormWithTranspose': [
                {
                    'patterns': {
                        'in': [[(0, 'MatMulWithBiasAdd'), (1, 'LayerNorm'), (2, 'Transpose')]],
                        'out': [[(0, 'MatMulWithBiasAdd'), (1, 'LayerNorm'), (2, 'Reshape'),
                                 (3, 'Transpose')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 1,
                        2: 'reshape_3d_before_transpose',
                        3: 2,
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            0: [1]
                        }, {
                            0: [2]
                        }, {
                            0: [3]
                        }], [[0, 1, 2, 3], 4]],
                        1: [[{
                            1: [1]
                        }, {
                            1: [2]
                        }], [[1, 2], 3]],
                        2: [[{
                            'input_data': [0]
                        }], [[1], 2]],
                        3: [[], [[], 1]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[], [[], 1]],
                        3: [[{
                            2: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 1, 2]
                },
            ]
        }

        def _set_attr(old_nodes, node_names, model):
            reshape_attr = OrderedDict()
            hidden_size = str(model.inquire_config_item("hidden_size"))
            reshape_attr['dst_shape'] = '-1,-1,' + hidden_size
            reshape_attr['dims'] = '1,0'

            reshape_idx = 0
            idx = 0
            for n in old_nodes:
                if model.get_node_by_name(node_names[idx]).op_type == "Reshape":
                    reshape_idx = idx
                    idx += 1
                model.get_node_by_name(new_node_names[j][idx]).attr = copy.deepcopy(n.attr)
                idx += 1
            model.get_node_by_name(new_node_names[j][reshape_idx]).attr = reshape_attr

        pattern_dict = pattern_mapping_config['LayerNormWithTranspose'][0]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("LayerNormWithTranspose",
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            for j in range(len(new_node_names)):
                _set_attr(ret_old_nodes[j], new_node_names[j], model)

            return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine tensorflow extractor file.

   Extract all nodes in the input tensorflow model and convert them to engine python operators.
   All these python operators will compose the engine graph in the order of the original tensorflow
   calculation graph. Please noticed that the mapping between operators is not one-to-one
   correspondence. For more related deatils, please refer to 'ops' APIs.
"""

from .. import logger
from ..graph.graph import Graph
from ..ops.op import OPERATORS
from ..tf_utils import graph_node_names_details
from ..graph_utils import names_from_input


class TensorflowExtractor(object):
    """The tensorflowExtractor Class.

    Decorate the node in model.graph_def, and the new node has the attributes like input_tensors
    and output_tensors, these tensors record the source/dest op name. All of these nodes
    (in a list) will compose a graph, which is Graph class, as the return object.
    
    Args:
        model: TensorflowModel

    Return:
        Graph: Graph class, the new graph object
    """
    @classmethod
    def __call__(self, model):
        """The __call__ function of the extractor."""
        nodes = model.graph_def.node
        graph_nodes_dict = graph_node_names_details(nodes)
        logger.info('Start to extarct tensorflow model ops...')
        new_graph = Graph()
        new_graph.framework_modeling_config['framework'] = 'tensorflow'
        for node in nodes:
            # ignore the Const nodes, they have no source inputs
            if node.op == 'Const':
                continue
            else:
                op_type = node.op
                if op_type not in OPERATORS.keys():
                    op_type = "OpAny"
                new_node = OPERATORS[op_type]()
                new_node.extract('tensorflow', node, model, graph_nodes_dict)
                new_graph.insert_nodes(len(new_graph.nodes), [new_node])

        return new_graph



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Length-Adaptive Transformer
# Copyright (c) 2020-present NAVER Corp.
# Apache License v2.0

"""Utils for drop and restore function."""

from dataclasses import dataclass, field
from typing import Optional, List

import numpy as np


def sample_length_configuration(
    max_seq_length,
    num_hidden_layers,
    layer_config=None,
    length_drop_prob=None,
    length_drop_ratio=None,
    length_drop_ratio_bound=None,
    min_length=2,
):
    """Get different sequence length for hidden layers.
    
    Args:
        max_seq_length: A number to set the max sequence length.
        num_hidden_layers: A number of total hidden layers.
        layer_config: Config to specify layers which use the max_seq_length.
        length_drop_prob: Probability to truncate the sequence.
        length_drop_ratio: Ratio to truncate the sequence.
        length_drop_ratio_bound: The max ratio to truncate the sequence.
            If the ratio set, the length will not less than max_seq_length * ratio.
        min_length: The number to set the min sequence length.
    
    Return:
        (Tuple): The tuple of length configuration for different hidden layers.
    """
    length = max_seq_length
    length_configuration = ()
    for i in range(num_hidden_layers):
        if layer_config is None or i in layer_config:
            if length_drop_prob is not None:
                length = length - np.random.binomial(length, length_drop_prob)
            elif length_drop_ratio is not None:
                length = int(np.ceil(length * (1 - length_drop_ratio)))
            elif length_drop_ratio_bound is not None:
                length = np.random.randint(int(np.ceil(length * (1 - length_drop_ratio_bound))), length + 1)
        length = max(length, min_length)
        length_configuration += (length,)
    return length_configuration


def sample_layer_configuration(
    num_hidden_layers,
    layer_dropout_prob=None,
    layer_dropout=None,
    layer_dropout_bound=None,
):
    """Get sample layers depends on the set paramaters.
    
    Args:
        num_hidden_layers: A number to set the max sequence length.
        layer_dropout_prob: Probability to dropout a layer.
        layer_dropout: Number of how many layers to dropout.
        layer_dropout_bound: The bound of how many layers to dropout.
    
    Return:
        (Tuple): The tuple to the numbers of which samples are kept.
    """
    if layer_dropout_prob is not None:
        return tuple(i for i in range(num_hidden_layers) if np.random.random() >= layer_dropout_prob)
    elif layer_dropout is not None:
        layer_dropout = min(layer_dropout, num_hidden_layers - 1)
        return tuple(range(num_hidden_layers - layer_dropout))
    elif layer_dropout_bound is not None:
        layer_dropout_bound = min(layer_dropout_bound, num_hidden_layers - 1)
        return tuple(range(num_hidden_layers - np.random.randint(0, layer_dropout_bound + 1)))
    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .op import Operator, operator_registry
from .tensor import Tensor
import numpy as np
from .. import logger


@operator_registry(operator_type=['Div', 'Sub', 'Mul', 'Equal', 'Less', 'Greater',
                                  'LessEqual', 'LessOrEqual', 'GreaterOrEqual',
                                  'NonZero', 'Not', 'Neg', 'NotEqual'])
class BinaryOp(Operator):
    def __init__(self):
        super().__init__()
        self._algorithm_dict = {'Addv2': 'add', 'Add': 'add', 'Div': 'div', 'Sub': 'sub',
                                'Mul': 'mul', 'Equal': 'eq', 'Less': 'lt', 'Greater': 'gt',
                                'LessEqual': 'le', 'LessOrEqual': 'le', 'GreaterOrEqual': 'ge',
                                'NonZero': 'ne', 'NotEqual': 'ne', 'Not': 'eq', 'Neg': 'mul'}

    def set_attr(self, framework, node):
        if framework == "onnxruntime":
            self._attr['algorithm'] = self._algorithm_dict[self._op_type]
            # deal with const input tensor
            if len(self._input_tensors) == 2:
                const_idx = -1
                for i in [0, 1]:
                    if isinstance(self._input_tensors[i].data, np.ndarray):
                        const_idx = i
                        break
                if const_idx != -1:
                    if len(self._input_tensors[const_idx].data.shape) == 0:
                        self._input_tensors[const_idx].shape = [1]
                        self._input_tensors[const_idx].data = \
                            self._input_tensors[const_idx].data.reshape(1)
                    # dnnl Binary op only supports f32, bf16, f16, u8, s8
                    # convert dtype int64 and int32 into dtype fp32
                    const_data = self._input_tensors[const_idx].data
                    if const_data.dtype in [np.int64, np.int32]:
                        if all(const_data >= np.finfo(np.float32).min) and \
                           all(const_data <= np.finfo(np.float64).max):
                            self._input_tensors[const_idx].data = const_data.astype(np.float32)
                        else:
                            logger.warning("Cannot convert const data into fp32 dtype in op type"\
                                          " {}".format(self._op_type))

            # convert other special ops to binary ops
            if self._op_type in ["NonZero", "Not"]:
                self._input_tensors.append(Tensor(name=self._name + "_equal_val",
                                                  data=np.array([0]).astype(np.float32),
                                                  shape=[1],
                                                  dest_op=[self._name])
                                           )
            if self._op_type in ["Neg"]:
                self._input_tensors.append(Tensor(name=self._name + "_mul_val",
                                                  data=np.array([-1]).astype(np.float32),
                                                  shape=[1],
                                                  dest_op=[self._name])
                                           )
        if framework == 'torch':
            algo_dict = {'aten::rsub': 'sub', 'aten::mul': 'mul', 'aten::add': 'add', 'aten::add_': 'add',
                         'aten::div': 'div', 'aten::sub': 'sub', 'aten::gt': 'gt', 'aten::lt': 'lt',
                         'aten::eq': 'eq', 'aten::ne': 'ne', 'aten::neg': 'mul'}
            self._attr['algorithm'] = algo_dict[node.kind()]
            if node.kind() == 'aten::neg':
                self._input_tensors.append(Tensor(name=self._name + "_mul_val",
                                                  data=np.array([-1]).astype(np.float32),
                                                  shape=[1],
                                                  dest_op=[self._name]))


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Config: provide config classes for optimization processes."""

import yaml
from enum import Enum
from neural_compressor.conf.config import (
    Distillation_Conf, Pruner, Pruning_Conf, Quantization_Conf
)
from neural_compressor.conf.dotdict import DotDict, deep_set
from intel_extension_for_transformers.optimization.utils.metrics import Metric
from intel_extension_for_transformers.optimization.utils.objectives import Objective, performance
from intel_extension_for_transformers.optimization.quantization import QuantizationMode, SUPPORTED_QUANT_MODE
from intel_extension_for_transformers.optimization.distillation import (
    Criterion, DistillationCriterionMode, SUPPORTED_DISTILLATION_CRITERION_MODE
)
from typing import List, Union
from xmlrpc.client import boolean


WEIGHTS_NAME = "pytorch_model.bin"


class Provider(Enum):
    """Optimization functionalities provider: INC or NNCF."""
    INC = "inc"


def check_value(name, src, supported_type, supported_value=[]):  # pragma: no cover
    """Check if the given object is the given supported type and in the given supported value.

    Example::

        def datatype(self, datatype):
            if check_value('datatype', datatype, list, ['fp32', 'bf16', 'uint8', 'int8']):
                self._datatype = datatype
    """
    if isinstance(src, list) and any([not isinstance(i, supported_type) for i in src]):
        assert False, ("Type of {} items should be {} but not {}".format(
            name, str(supported_type), [type(i) for i in src]))
    elif not isinstance(src, list) and not isinstance(src, supported_type):
        assert False, ("Type of {} should be {} but not {}".format(
            name, str(supported_type), type(src)))

    if len(supported_value) > 0:
        if isinstance(src, str) and src not in supported_value:
            assert False, ("{} is not in supported {}: {}. Skip setting it.".format(
                src, name, str(supported_value)))
        elif isinstance(src, list) and all([isinstance(i, str) for i in src]) and \
            any([i not in supported_value for i in src]):
            assert False, ("{} is not in supported {}: {}. Skip setting it.".format(
                src, name, str(supported_value)))

    return True
def constructor_register(cls):
    yaml_key = "!{}".format(cls.__name__)

    def constructor(loader, node):
        instance = cls.__new__(cls)
        yield instance

        state = loader.construct_mapping(node, deep=True)
        instance.__init__(**state)

    yaml.add_constructor(
        yaml_key,
        constructor,
        yaml.SafeLoader,
    )
    return cls


class DynamicLengthConfig(object):
    """Configure the dynamic length config for Quantized Length Adaptive Transformer.

    Args:
        max_length: Limit the maximum length of each layer
        length_config: The length number for each layer
        const_rate: Length drop ratio
        num_sandwich: Sandwich num used in training
        length_drop_ratio_bound: Length dropout ratio list
        layer_dropout_prob: The layer dropout with probability
        layer_dropout_bound: Length dropout ratio
        dynamic_training: Whether to use dynamic training
        load_store_file: The path for store file
        evo_iter: Iterations for evolution search
        population_size: Population limitation for evolution search
        mutation_size: Mutation limitation for evolution search
        mutation_prob: Mutation probability used in evolution search
        crossover_size: Crossover limitation for evolution search
        num_cpus: The cpu nums used in evolution search
        distributed_world_size: Distributed world size in evolution search training
        latency_constraint: Latency constraint used in evolution search
        evo_eval_metric: The metric name used in evolution search
    """
    def __init__(
        self,
        max_length: int = None,
        length_config: str = None,
        const_rate: float = None,
        num_sandwich: int = 2,
        length_drop_ratio_bound: float = 0.2,
        layer_dropout_prob: float = None,
        layer_dropout_bound: int = 0,
        dynamic_training: bool = False,
        load_store_file: str = None,
        evo_iter: int = 30,
        population_size: int = 20,
        mutation_size: int = 30,
        mutation_prob: float = 0.5,
        crossover_size: int = 30,
        num_cpus: int = 48,
        distributed_world_size: int = 5,
        latency_constraint: bool = True,
        evo_eval_metric = 'eval_f1'
    ):
        """Init a DynamicLengthConfig object."""
        super().__init__()

        self.length_config = length_config
        self.const_rate = const_rate
        self.max_length = max_length
        self.num_sandwich = num_sandwich
        self.length_drop_ratio_bound = length_drop_ratio_bound
        self.layer_dropout_prob = layer_dropout_prob
        self.layer_dropout_bound = layer_dropout_bound
        self.dynamic_training = dynamic_training
        self.load_store_file = load_store_file
        self.evo_iter = evo_iter
        self.population_size = population_size
        self.mutation_size = mutation_size
        self.mutation_prob = mutation_prob
        self.crossover_size = crossover_size
        self.num_cpus = num_cpus
        self.distributed_world_size = distributed_world_size
        self.latency_constraint = latency_constraint
        self.evo_eval_metric = evo_eval_metric


class QuantizationConfig(object):
    """Configure the quantization process.

    Args:
        framework: Which framework you used
        approach: Which quantization approach to use
        timeout: Tuning timeout(seconds), 0 means early stop. Combined with max_trials field to decide when to exit
        max_trials: Max tune times
        metrics: Used to evaluate accuracy of tuning model, no need for NoTrainerOptimize
        objectives: Objective with accuracy constraint guaranteed
        config_file: Path to the config file
        sampling_size: How many samples to use
        use_bf16: Whether to use bf16
        recipes: apply recipes for quantization, neural_compressor support below recipes:
            'smooth_quant': whether do smooth quant
            'smooth_quant_args': parameters for smooth_quant
            'fast_bias_correction': whether do fast bias correction
            'weight_correction': whether do weight correction
            'gemm_to_matmul': whether convert gemm to matmul and add, only valid for onnx models
            'graph_optimization_level': support 'DISABLE_ALL', 'ENABLE_BASIC', 'ENABLE_EXTENDED', 'ENABLE_ALL'
                                      only valid for onnx models
            'first_conv_or_matmul_quantization': whether quantize the first conv or matmul
            'last_conv_or_matmul_quantization': whether quantize the last conv or matmul
            'pre_post_process_quantization': whether quantize the ops in preprocess and postprocess
            'add_qdq_pair_to_weight': whether add QDQ pair for weights, only vaild for onnxrt_trt_ep
            'optypes_to_exclude_output_quant': don't quantize output of specified optypes
            'dedicated_qdq_pair': whether dedicate QDQ pair, only vaild for onnxrt_trt_ep.
    """
    def __init__(
        self,
        framework: str = "pytorch",
        approach: str = "PostTrainingStatic",
        timeout: int = 0,
        max_trials: int = 100,
        metrics: Union[Metric, List] = None,
        objectives: Union[Objective, List] = performance,
        config_file: str = None,
        sampling_size: int = 100,
        use_bf16: bool = False,
        recipes: dict = None,
    ):
        """Init a QuantizationConfig object."""
        super().__init__()
        if config_file is None:
            self.inc_config = Quantization_Conf()
        else:
            self.inc_config = Quantization_Conf(config_file)
        self.framework = framework
        if approach is not None:
            self.approach = approach
        if timeout is not None:
            self.timeout = timeout
        if max_trials is not None:
            self.max_trials = max_trials
        if metrics is not None:
            self.metrics = metrics
        else:
            self._metrics = None
        if objectives is not None:
            self.objectives = objectives
        else:
            self._objectives = None
        if sampling_size is not None:
            self.sampling_size = sampling_size
        self.inc_config.usr_cfg.use_bf16 = use_bf16
        if recipes is not None:
            self.recipes = recipes

    @property
    def approach(self):
        """Get the quantization approach."""
        return self.inc_config.usr_cfg.quantization.approach

    @approach.setter
    def approach(self, approach):
        """Set the quantization approach."""
        approach = approach.upper()
        assert approach in SUPPORTED_QUANT_MODE, \
            f"quantization approach: {approach} is not support!" + \
            "PostTrainingStatic, PostTrainingDynamic and QuantizationAwareTraining are supported!"
        self.inc_config.usr_cfg.quantization.approach = QuantizationMode[approach].value

    @property
    def input_names(self):
        """Get the input names."""
        return self.inc_config.usr_cfg.model.inputs

    @input_names.setter
    def input_names(self, input_names):
        """Set the input names."""
        assert isinstance(input_names, list), "input_names must be a list"
        self.inc_config.usr_cfg.model.inputs = input_names

    @property
    def output_names(self):
        """Get the output names."""
        return self.inc_config.usr_cfg.model.outputs

    @output_names.setter
    def output_names(self, output_names):
        """Set the output names."""
        assert isinstance(output_names, list), "output_names must be a list"
        self.inc_config.usr_cfg.model.outputs = output_names

    @property
    def metrics(self):
        """Get the metrics."""
        return self._metrics

    @metrics.setter
    def metrics(self, metrics: Union[Metric, List]):
        """Set the metrics."""
        self._metrics = metrics
        rel_or_abs = {True: "relative", False: "absolute"}
        assert isinstance(metrics[0] if isinstance(metrics, list) else metrics, Metric), \
            "metric should be a Metric calss!"
        if isinstance(metrics, Metric) or len(metrics) == 1:
            self.inc_config.usr_cfg.tuning.accuracy_criterion = {
                rel_or_abs[metrics[0].is_relative]
                if isinstance(metrics, list) else rel_or_abs[metrics.is_relative]:
                metrics[0].criterion if isinstance(metrics, list) else metrics.criterion,
                "higher_is_better": metrics[0].greater_is_better if isinstance(metrics, list) else
                metrics.greater_is_better
            }
        else:
            weights = [metric.weight_ratio for metric in metrics]
            if not any(weights):
                weight = 1 / len(metrics)
                for metric in metrics:
                    metric.weight_ratio = weight
            else:   # pragma: no cover
                assert all(weights), "Please set the weight ratio for all metrics!"

            assert all(metric.is_relative == metrics[0].is_relative for metric in metrics), \
                "Unsupport different is_relative for different metric now, will support soon!"
            assert all(metric.criterion == metrics[0].criterion for metric in metrics), \
                "Unsupport different criterion for different metric now, will support soon!"

            self.inc_config.usr_cfg.tuning.accuracy_criterion = {
                rel_or_abs[metrics[0].is_relative]: metrics[0].criterion,
                "higher_is_better": metrics[0].greater_is_better
            }

    @property
    def framework(self):
        """Get the framework."""
        return self.inc_config.usr_cfg.model.framework

    @framework.setter
    def framework(self, framework):
        """Set the framework."""
        assert framework in ["pytorch", "pytorch_fx", "pytorch_ipex", "tensorflow"], \
            "framework: {} is not support!".format(framework)
        self.inc_config.usr_cfg.model.framework = framework

    @property
    def objectives(self):
        """Get the objectives."""
        return self._objectives

    @objectives.setter
    def objectives(self, objectives: Union[List, Objective]):
        """Set the objectives."""
        self._objectives = objectives
        if isinstance(objectives, Objective) or len(objectives) == 1:
            self.inc_config.usr_cfg.tuning.objective = objectives.name \
                if isinstance(objectives, Objective) else objectives[0].name
        else:
            weights = [objective.weight_ratio for objective in objectives]
            if not any(weights):
                weight = 1 / len(objectives)
                for objective in objectives:
                    objective.weight_ratio = weight
            else:
                assert all(weights), "Please set the weight ratio for all metrics!"

            self.inc_config.usr_cfg.tuning.multi_objective = {
                "objective": [objective.name for objective in objectives],
                "higher_is_better": [objective.greater_is_better for objective in objectives],
                "weight": [objective.weight_ratio for objective in objectives],
            }

    @property
    def strategy(self):
        """Get the strategy."""
        return self.inc_config.usr_cfg.tuning.strategy.name

    @strategy.setter
    def strategy(self, strategy):
        """Set the strategy."""
        assert strategy in ["basic", "bayesian", "mse", "mse_v2"], \
            "strategy: {} is not support!".format(strategy)
        self.inc_config.usr_cfg.tuning.strategy.name = strategy
        if strategy == "mse_v2":
                 self.inc_config.usr_cfg.tuning.strategy_kwargs = {"confidence_batches": 1}

    @property
    def timeout(self):
        """Get the timeout."""
        return self.inc_config.usr_cfg.tuning.exit_policy.timeout

    @timeout.setter
    def timeout(self, timeout):
        """Set the timeout."""
        assert isinstance(timeout, int), "timeout should be integer!"
        self.inc_config.usr_cfg.tuning.exit_policy.timeout = timeout

    @property
    def op_wise(self):
        """Get the op_wise dict."""
        return self.inc_config.usr_cfg.quantization.op_wise

    @op_wise.setter
    def op_wise(self, op_wise):
        """Set the op_wise dict."""
        self.inc_config.usr_cfg.quantization.op_wise = op_wise

    @property
    def max_trials(self):
        """Get the number of maximum trials."""
        return self.inc_config.usr_cfg.tuning.exit_policy.max_trials

    @max_trials.setter
    def max_trials(self, max_trials):
        """Set the number of maximum trials."""
        assert isinstance(max_trials, int), "max_trials should be integer!"
        self.inc_config.usr_cfg.tuning.exit_policy.max_trials = max_trials

    @property
    def performance_only(self):
        """Get the boolean whether to use performance only."""
        return self.inc_config.usr_cfg.tuning.exit_policy.performance_only

    @performance_only.setter
    def performance_only(self, performance_only):
        """Set the boolean whether to use performance only."""
        assert isinstance(performance_only, boolean), "performance_only should be boolean!"
        self.inc_config.usr_cfg.tuning.exit_policy.performance_only = performance_only

    @property
    def random_seed(self):
        """Get the random seed."""
        return self.inc_config.usr_cfg.tuning.random_seed

    @random_seed.setter
    def random_seed(self, random_seed):
        """Set the random seed."""
        assert isinstance(random_seed, int), "random_seed should be integer!"
        self.inc_config.usr_cfg.tuning.random_seed = random_seed

    @property
    def tensorboard(self):
        """Get the boolean whether to use tensorboard."""
        return self.inc_config.usr_cfg.tuning.tensorboard

    @tensorboard.setter
    def tensorboard(self, tensorboard):
        """Set the boolean whether to use tensorboard."""
        assert isinstance(tensorboard, boolean), "tensorboard should be boolean!"
        self.inc_config.usr_cfg.tuning.tensorboard = tensorboard

    @property
    def output_dir(self):
        """Get the output directory."""
        return self.inc_config.usr_cfg.tuning.workspace.path

    @output_dir.setter
    def output_dir(self, path):
        """Set the output directory."""
        assert isinstance(path, str), "save_path should be a string of directory!"
        self.inc_config.usr_cfg.tuning.workspace.path = path

    @property
    def resume_path(self):
        """Get the resume path."""
        return self.inc_config.usr_cfg.tuning.workspace.resume

    @resume_path.setter
    def resume_path(self, path):
        """Set the resume path."""
        assert isinstance(path, str), "resume_path should be a string of directory!"
        self.inc_config.usr_cfg.tuning.workspace.resume = path

    @property
    def sampling_size(self):
        """Get the sampling size."""
        return self.inc_config.usr_cfg.quantization.calibration.sampling_size

    @sampling_size.setter
    def sampling_size(self, sampling_size):
        """Set the sampling size."""
        if isinstance(sampling_size, int):
            self.inc_config.usr_cfg.quantization.calibration.sampling_size = [sampling_size]
        elif isinstance(sampling_size, list):
            self.inc_config.usr_cfg.quantization.calibration.sampling_size = sampling_size
        else:
            assert False, "The sampling_size must be a list of int numbers"

    @property
    def recipes(self):
        """Get the sampling size."""
        return self.inc_config.usr_cfg.quantization.recipes

    @recipes.setter
    def recipes(self, recipes):
        """Set recipes."""
        if recipes is not None and not isinstance(recipes, dict):
            raise ValueError("recipes should be a dict.")

        # Support PyTorch only
        def smooth_quant(val=None):
            if val is not None:
                return check_value("smooth_quant", val, bool)
            else:
                return False

        # Support PyTorch only
        def smooth_quant_args(val=None):
            if val is not None:
                check_value("smooth_quant_args", val, dict)
                for k, v in val.items():
                    if k == "alpha":
                        assert isinstance(v, str) or isinstance(v, float),\
                            "Smooth_quant_args.alpha should be a float or 'auto'."
                return True
            else:
                return {}

        # Support tensorflow, but not enabled now
        def fast_bias_correction(val=None):  # pragma: no cover
            if val is not None:
                return check_value("fast_bias_correction", val, bool)
            else:
                return False

        # Support tensorflow, but not enabled now
        def weight_correction(val=None):  # pragma: no cover
            if val is not None:
                return check_value("weight_correction", val, bool)
            else:
                return False

        # Support Tensorflow only
        def first_conv_or_matmul_quantization(val=None):
            if val is not None:
                return check_value("first_conv_or_matmul_quantization", val, bool)
            else:
                return True

        # Support Tensorflow only
        def last_conv_or_matmul_quantization(val=None):
            if val is not None:
                return check_value("last_conv_or_matmul_quantization", val, bool)
            else:
                return True

        RECIPES = {"smooth_quant": smooth_quant, # Only for PyTorch
                   "smooth_quant_args": smooth_quant_args, # Only for PyTorch
                   "fast_bias_correction": fast_bias_correction, # Support PyTorch and Tensorflow, not used now.
                   "weight_correction": weight_correction, # Support PyTorch and Tensorflow, not used now.
                   "first_conv_or_matmul_quantization": first_conv_or_matmul_quantization, # Only for Tensorflow
                   "last_conv_or_matmul_quantization": last_conv_or_matmul_quantization, # Only for Tensorflow
                   }
        _recipes = {}
        for k in RECIPES.keys():
            if k in recipes and RECIPES[k](recipes[k]):
                _recipes.update({k: recipes[k]})
            else:
                _recipes.update({k: RECIPES[k]()})
        deep_set(self.inc_config.usr_cfg, 'quantization.recipes', _recipes)


class PruningConfig(object):
    """Configure the pruning process.

    Args:
        framework: Which framework you used
        epochs: How many epochs to prune
        epoch_range: Epoch range list
        initial_sparsity_ratio: Initial sparsity goal, and not needed if pruner_config argument is defined
        target_sparsity_ratio: Target sparsity goal, and not needed if pruner_config argument is defined
        metrics: Used to evaluate accuracy of tuning model, not needed for NoTrainerOptimizer
        pruner_config: Defined pruning behavior, if it is None, then NLP wil create a default pruner with
            'BasicMagnitude' pruning typel
        config_file: Path to the config file
    """
    def __init__(
        self,
        framework: str = "pytorch",
        epochs: int = 1,
        epoch_range: List = [0, 4],
        initial_sparsity_ratio: float=0.0,
        target_sparsity_ratio: float = 0.97,
        metrics: Metric = None,
        pruner_config: Union[List, Pruner] = None,
        config_file: str = None
    ):
        """Init a PruningConfig object."""
        super().__init__()
        self.inc_config = Pruning_Conf(config_file)
        self.framework = framework

        if initial_sparsity_ratio is not None:
            self.initial_sparsity_ratio = initial_sparsity_ratio
        if target_sparsity_ratio is not None:
            self.target_sparsity_ratio = target_sparsity_ratio
        if epoch_range is not None:
            self.epoch_range = epoch_range
        if metrics is not None:
            self.metrics = metrics
        else:
            self._metrics = None
        if pruner_config is not None:
            self.pruner_config = pruner_config
        else:
            self.init_prune_config()
        self.epochs = epochs


    def init_prune_config(self):
        """Init the pruning config."""
        pruner_config = Pruner()
        self.inc_config.usr_cfg.pruning.approach.weight_compression['pruners'] = [pruner_config]

    @property
    def pruner_config(self):
        """Get the pruner config."""
        return self.inc_config.usr_cfg.pruning.approach.weight_compression.pruners

    @pruner_config.setter
    def pruner_config(self, pruner_config):
        """Set the pruner config."""
        if isinstance(pruner_config, list):
            self.inc_config.usr_cfg.pruning.approach.weight_compression.pruners = pruner_config
        else:
            self.inc_config.usr_cfg.pruning.approach.weight_compression.pruners = [pruner_config]

    @property
    def target_sparsity_ratio(self):
        """Get the target sparsity ratio."""
        return self.inc_config.usr_cfg.pruning.approach.weight_compression.target_sparsity

    @target_sparsity_ratio.setter
    def target_sparsity_ratio(self, target_sparsity_ratio):
        """Set the target sparsity ratio."""
        self.inc_config.usr_cfg.pruning.approach.weight_compression.target_sparsity = \
            target_sparsity_ratio

    @property
    def initial_sparsity_ratio(self):
        """Get the initial sparsity ratio."""
        return self.inc_config.usr_cfg.pruning.approach.weight_compression.initial_sparsity

    @initial_sparsity_ratio.setter
    def initial_sparsity_ratio(self, initial_sparsity_ratio):
        """Set the initial sparsity ratio."""
        self.inc_config.usr_cfg.pruning.approach.weight_compression.initial_sparsity = \
            initial_sparsity_ratio

    @property
    def epoch_range(self):
        """Get the epoch range."""
        return [self.inc_config.usr_cfg.pruning.approach.weight_compression.start_epoch,
                self.inc_config.usr_cfg.pruning.approach.weight_compression.end_epoch]

    @epoch_range.setter
    def epoch_range(self, epoch_range):
        """Set the epoch range."""
        assert isinstance(epoch_range, list) and len(epoch_range) == 2, \
          "You should set epoch_range like [a,b] format to match the pruning start and end epoch."
        self.inc_config.usr_cfg.pruning.approach.weight_compression.start_epoch = epoch_range[0]
        self.inc_config.usr_cfg.pruning.approach.weight_compression.end_epoch = epoch_range[1]

    @property
    def epochs(self):
        """Get the epochs."""
        eps = self.inc_config.usr_cfg.pruning.train.epoch \
            if hasattr(self.inc_config.usr_cfg.pruning, "train") else 1
        return eps

    @epochs.setter
    def epochs(self, epochs):
        """Set the epochs."""
        assert isinstance(epochs, int) and epochs > 0, \
          "You should set epochs > 0 and int, not {}.".format(epochs)
        self.inc_config.usr_cfg.pruning["train"] = {"epoch": epochs}

    @property
    def framework(self):
        """Get the framework."""
        return self.inc_config.usr_cfg.model.framework

    @framework.setter
    def framework(self, framework):
        """Set the framework."""
        assert framework.lower() in ["pytorch", "pytorch_fx", "tensorflow"], \
            "framework: {} is not support!".format(framework)
        self.inc_config.usr_cfg.model.framework = framework.lower()

    @property
    def metrics(self):
        """Get the metrics."""
        return self._metrics

    @metrics.setter
    def metrics(self, metrics: Metric):
        """Set the metrics."""
        self._metrics = metrics


class DistillationConfig(object):
    """Configure the distillation process.

    Args:
        framework: Which framework you used
        criterion: Criterion of training, example: "KnowledgeLoss"
        metrics: Metrics for distillation
        inc_config: Distillation config
    """
    def __init__(
        self,
        framework: str = "pytorch",
        criterion: Criterion = None,
        metrics: Metric = None,
        inc_config = None
    ):
        """Init a DistillationConfig object."""
        super().__init__()
        self.inc_config = Distillation_Conf(inc_config)
        self.framework = framework
        if criterion is not None:
            self.criterion = criterion
        if metrics is not None:
            self.metrics = metrics
        else:
            self._metrics = None

    @property
    def framework(self):
        """Get the framework."""
        return self.inc_config.usr_cfg.model.framework

    @framework.setter
    def framework(self, framework):
        """Set the framework."""
        assert framework in ["pytorch", "pytorch_fx", "tensorflow"], \
            "framework: {} is not support!".format(framework)
        self.inc_config.usr_cfg.model.framework = framework

    @property
    def criterion(self):
        """Get the criterion."""
        return self.inc_config.usr_cfg.distillation.train.criterion

    @criterion.setter
    def criterion(self, criterion: Criterion):
        """Set the criterion."""
        assert criterion.name.upper() in SUPPORTED_DISTILLATION_CRITERION_MODE, \
            "The criterion name must be in ['KnowledgeLoss', 'IntermediateLayersLoss']"
        if criterion.name.upper() == DistillationCriterionMode.KNOWLEDGELOSS.name:
            assert criterion.temperature is not None, \
                "Please pass the temperature to Criterion.temperature!"
            assert criterion.loss_types is not None, \
                "Please pass the loss_types to Criterion.loss_types!"
            assert criterion.loss_weight_ratio is not None, \
                "Please pass the loss_weight_ratio to Criterion.loss_weight_ratio!"
            self.inc_config.usr_cfg.distillation.train.criterion = {
                DistillationCriterionMode.KNOWLEDGELOSS.value: {
                    "temperature": criterion.temperature,
                    "loss_types": criterion.loss_types,
                    "loss_weights": criterion.loss_weight_ratio
                }
            }

        if criterion.name.upper() == DistillationCriterionMode.INTERMEDIATELAYERSLOSS.name:
            assert criterion.layer_mappings is not None, \
                "Please pass the layer_mappings to Criterion.layer_mappings!"
            assert criterion.loss_types is not None, \
                "Please pass the loss_types to Criterion.loss_types!"
            assert criterion.loss_weight_ratio is not None, \
                "Please pass the loss_weight_ratio to Criterion.loss_weight_ratio!"
            assert criterion.add_origin_loss is not None, \
                "Please pass the add_origin_loss to Criterion.add_origin_loss!"
            self.inc_config.usr_cfg.distillation.train.criterion = {
                DistillationCriterionMode.INTERMEDIATELAYERSLOSS.value: {
                    "layer_mappings": criterion.layer_mappings,
                    "loss_types": criterion.loss_types,
                    "loss_weights": criterion.loss_weight_ratio,
                    "add_origin_loss": criterion.add_origin_loss
                }
            }

    @property
    def metrics(self):
        """Get the metrics."""
        return self._metrics

    @metrics.setter
    def metrics(self, metrics):
        """Set the metrics."""
        assert isinstance(metrics, Metric), \
            "metric should be a Metric calss!"
        self._metrics = metrics


class TFDistillationConfig(object):
    """Configure the distillation process for Tensorflow.

    Args:
        loss_types: Type of loss
        loss_weights: Weight ratio of loss
        train_steps: Steps of training
        temperature: Parameter for KnowledgeDistillationLoss
    """
    def __init__(
        self,
        loss_types: list = [],
        loss_weights: list = [],
        train_steps: list = [],
        temperature: float = 1.0
    ):
        """Init a TFDistillationConfig object."""
        super().__init__()
        self.loss_types = loss_types
        self.loss_weights = loss_weights
        self.train_steps = train_steps
        self.temperature = temperature


class FlashDistillationConfig(object):
    """The flash distillation configuration used by AutoDistillationConfig."""
    def __init__(
        self,
        block_names: list = [],
        layer_mappings_for_knowledge_transfer: list = [],
        loss_types: list = [],
        loss_weights: list = [],
        add_origin_loss: list = [],
        train_steps: list = [],
    ):
        """Init a FlashDistillationConfig object."""
        super().__init__()
        self.block_names = block_names
        self.layer_mappings_for_knowledge_transfer = layer_mappings_for_knowledge_transfer
        self.loss_types = loss_types
        self.loss_weights = loss_weights
        self.add_origin_loss = add_origin_loss
        self.train_steps = train_steps


class AutoDistillationConfig(object):
    """Configure the auto disillation process.

    Args:
        framework: Which framework you used
        search_space: Search space of NAS
        search_algorithm: Search algorithm used in NAS, e.g. Bayesian Optimization
        metrics: Metrics used to evaluate the performance of the model architecture candidate
        max_trials: Maximum trials in NAS process
        seed: Seed of random process
        knowledge_transfer: Configuration controlling the behavior of knowledge transfer stage
            in the autodistillation
        regular_distillation: Configuration controlling the behavior of regular distillation stage
            in the autodistillation
    """
    def __init__(
        self,
        framework: str = "pytorch",
        search_space: dict = {},
        search_algorithm: str = "BO",
        metrics: Union[List, Metric] = None,
        max_trials: int = None,
        seed: int = None,
        knowledge_transfer: FlashDistillationConfig = None,
        regular_distillation: FlashDistillationConfig = None,
    ):
        """Init a AutoDistillationConfig object."""
        super().__init__()
        self.config = DotDict({
            'model':{'name': 'AutoDistillation'},
            'auto_distillation':{
                'search':{},
                'flash_distillation':{
                    'knowledge_transfer':{},
                    'regular_distillation':{}
                    }
                }
            }
        )
        self.framework = framework
        self.search_space = search_space
        self.search_algorithm = search_algorithm
        if max_trials is not None:
            self.max_trials = max_trials
        if seed is not None:
            self.seed = seed
        if knowledge_transfer is not None:
            self.knowledge_transfer = knowledge_transfer
        if regular_distillation is not None:
            self.regular_distillation = regular_distillation
        if metrics is not None:
            self.metrics = metrics

    @property
    def knowledge_transfer(self):
        """Get the knowledge transfer."""
        return self.config.auto_distillation.flash_distillation.knowledge_transfer

    @knowledge_transfer.setter
    def knowledge_transfer(self, knowledge_transfer):
        """Set the knowledge transfer."""
        if knowledge_transfer is None:
            knowledge_transfer = FlashDistillationConfig()
        for k, v in knowledge_transfer.__dict__.items():
            if v:
                self.config.auto_distillation.flash_distillation.knowledge_transfer[k] = v

    @property
    def regular_distillation(self):
        """Get the regular distillation."""
        return self.config.auto_distillation.flash_distillation.regular_distillation

    @regular_distillation.setter
    def regular_distillation(self, regular_distillation):
        """Set the regular distillation."""
        if regular_distillation is None:
            regular_distillation = FlashDistillationConfig()
        for k, v in regular_distillation.__dict__.items():
            if v:
                self.config.auto_distillation.flash_distillation.regular_distillation[k] = v

    @property
    def framework(self):
        """Get the framework."""
        return self.config.model.framework

    @framework.setter
    def framework(self, framework):
        """Set the framework."""
        assert framework in ["pytorch"], \
            "framework: {} is not support!".format(framework)
        self.config.model.framework = framework

    @property
    def search_space(self):
        """Get the search space."""
        return self.config.auto_distillation.search.search_space

    @search_space.setter
    def search_space(self, search_space: dict):
        """Set the search space."""
        self.config.auto_distillation.search.search_space = search_space

    @property
    def search_algorithm(self):
        """Get the search algorithm."""
        return self.config.auto_distillation.search.search_algorithm

    @search_algorithm.setter
    def search_algorithm(self, search_algorithm: str):
        """Set the search algorithm."""
        self.config.auto_distillation.search.search_algorithm = search_algorithm

    @property
    def max_trials(self):
        """Get the max trials."""
        return self.config.auto_distillation.search.max_trials

    @max_trials.setter
    def max_trials(self, max_trials: int):
        """Set the max trials."""
        self.config.auto_distillation.search.max_trials = max_trials

    @property
    def seed(self):
        """Get the seed."""
        return self.config.auto_distillation.search.seed

    @seed.setter
    def seed(self, seed: int):
        """Set the seed."""
        self.config.auto_distillation.search.seed = seed

    @property
    def metrics(self):
        """Get the metrics."""
        return self._metrics

    @metrics.setter
    def metrics(self, metrics: Union[List, Metric]):
        """Set the metrics."""
        self._metrics = metrics
        if isinstance(metrics, Metric):
            metrics = [metrics]
        self.config.auto_distillation.search.metrics = []
        self.config.auto_distillation.search.higher_is_better = []
        for metric in metrics:
            self.config.auto_distillation.search.metrics.append(metric.name)
            self.config.auto_distillation.search.higher_is_better.append(
                metric.greater_is_better
                )


class NASConfig(object):
    """config parser.

    Args:
        approach: The approach of the NAS.
        search_algorithm: The search algorithm for NAS procedure.

    """

    def __init__(self,
        framework: str = "pytorch",
        approach: str = "basic",
        search_space: dict = {},
        search_algorithm: str = "BO",
        metrics: Union[List, Metric] = None,
        max_trials: int = None,
        seed: int = None,
        ):
        super().__init__()
        self.config = DotDict({
                'model': {'name': 'nas', 'framework': 'NA'},
                'nas': {
                    'approach': approach,
                    'search': {
                        'search_space': search_space,
                        'search_algorithm': search_algorithm
                    }
                },
            }
        )
        self.framework = framework
        self.search_space = search_space
        self.search_algorithm = search_algorithm
        if max_trials is not None:
            self.max_trials = max_trials
        if seed is not None:
            self.seed = seed
        if metrics is not None:
            self.metrics = metrics
        if approach and approach != 'basic':
            self.config[approach] = DotDict({})
            self.__setattr__(approach, self.config[approach])

    @property
    def framework(self):
        """Get the framework."""
        return self.config.model.framework

    @framework.setter
    def framework(self, framework):
        """Set the framework."""
        assert framework in ["pytorch"], \
            "framework: {} is not support!".format(framework)
        self.config.model.framework = framework

    @property
    def search_space(self):
        """Get the search space."""
        return self.config.nas.search.search_space

    @search_space.setter
    def search_space(self, search_space: dict):
        """Set the search space."""
        self.config.nas.search.search_space = search_space

    @property
    def search_algorithm(self):
        """Get the search algorithm."""
        return self.config.nas.search.search_algorithm

    @search_algorithm.setter
    def search_algorithm(self, search_algorithm: str):
        """Set the search algorithm."""
        self.config.nas.search.search_algorithm = search_algorithm

    @property
    def max_trials(self):
        """Get the max trials."""
        return self.config.nas.search.max_trials

    @max_trials.setter
    def max_trials(self, max_trials: int):
        """Set the max trials."""
        self.config.nas.search.max_trials = max_trials

    @property
    def seed(self):
        """Get the seed."""
        return self.config.nas.search.seed

    @seed.setter
    def seed(self, seed: int):
        """Set the seed."""
        self.config.nas.search.seed = seed

    @property
    def metrics(self):
        """Get the metrics."""
        return self._metrics

    @metrics.setter
    def metrics(self, metrics: Union[List, Metric]):
        """Set the metrics."""
        self._metrics = metrics
        if isinstance(metrics, Metric):
            metrics = [metrics]
        self.config.nas.search.metrics = []
        self.config.nas.search.higher_is_better = []
        for metric in metrics:
            self.config.nas.search.metrics.append(metric.name)
            self.config.nas.search.higher_is_better.append(
                metric.greater_is_better
                )

class BenchmarkConfig:
    """Config Class for Benchmark.

    Args:
        backend (str, optional): the backend used for benchmark. Defaults to "torch".
        warmup (int, optional): skip iters when collecting latency. Defaults to 5.
        iteration (int, optional): total iters when collecting latency. Defaults to 20.
        cores_per_instance (int, optional): the core number for 1 instance. Defaults to 4.
        num_of_instance (int, optional): the instance number. Defaults to -1.
        torchscript (bool, optional): Enable it if you want to jit trace it \
                                      before benchmarking. Defaults to False.
        generate (bool, optional): Enable it if you want to use model.generate \
                                   when benchmarking. Defaults to False.
    """
    def __init__(
        self,
        backend: str = "torch",  # select from ["torch", "ipex", "neural_engine"]
        batch_size: int = 1,
        warmup: int = 5,
        iteration: int = 20,
        cores_per_instance: int = 4,
        num_of_instance: int = -1,
        torchscript: bool = False,
        generate: bool = False,
        **kwargs,
    ):
        """Init a BenchmarkConfig object."""
        self.backend = backend
        self.batch_size = batch_size
        self.warmup = warmup
        self.iteration = iteration
        self.cores_per_instance = cores_per_instance
        self.num_of_instance = num_of_instance
        self.torchscript = torchscript
        self.generate = generate
        self.kwargs = kwargs

    @property
    def backend(self):
        """Get backend."""
        return self._backend

    @backend.setter
    def backend(self, backend):
        """Set backend."""
        self._backend = backend

    @property
    def batch_size(self):
        """Get batch_size."""
        return self._batch_size

    @batch_size.setter
    def batch_size(self, batch_size):
        """Set batch_size."""
        self._batch_size = batch_size

    @property
    def warmup(self):
        """Get warmup."""
        return self._warmup

    @warmup.setter
    def warmup(self, warmup):
        """Set warmup."""
        self._warmup = warmup

    @property
    def iteration(self):
        """Get iteration."""
        return self._iteration

    @iteration.setter
    def iteration(self, iteration):
        """Set iteration."""
        self._iteration = iteration

    @property
    def cores_per_instance(self):
        """Get cores_per_instance."""
        return self._cores_per_instance

    @cores_per_instance.setter
    def cores_per_instance(self, cores_per_instance):
        """Set cores_per_instance."""
        self._cores_per_instance = cores_per_instance

    @property
    def num_of_instance(self):
        """Get num_of_instance."""
        return self._num_of_instance

    @num_of_instance.setter
    def num_of_instance(self, num_of_instance):
        """Set num_of_instance."""
        self._num_of_instance = num_of_instance

    @property
    def torchscript(self):
        """Get torchscript."""
        return self._torchscript

    @torchscript.setter
    def torchscript(self, torchscript):
        """Set torchscript."""
        self._torchscript = torchscript

    @property
    def generate(self):
        """Get generate."""
        return self._generate

    @generate.setter
    def generate(self, generate):
        """Set generate."""
        self._generate = generate

    @property
    def kwargs(self):
        """Get kwargs."""
        return self._kwargs

    @kwargs.setter
    def kwargs(self, kwargs):
        """Set kwargs."""
        self._kwargs = kwargs
        
@constructor_register
class PrunerV2:
    """
    similiar to torch optimizer's interface
    """

    def __init__(self,
                 target_sparsity=None, pruning_type=None, pattern=None, op_names=None,
                 excluded_op_names=None,
                 start_step=None, end_step=None, pruning_scope=None, pruning_frequency=None,
                 min_sparsity_ratio_per_op=None, max_sparsity_ratio_per_op=None,
                 sparsity_decay_type=None, pruning_op_types=None, reg_type=None,
                 criterion_reduce_type=None, parameters=None, resume_from_pruned_checkpoint=None):
        self.pruner_config = DotDict({
            'target_sparsity': target_sparsity,
            'pruning_type': pruning_type,
            'pattern': pattern,
            'op_names': op_names,
            'excluded_op_names': excluded_op_names,  ##global only
            'start_step': start_step,
            'end_step': end_step,
            'pruning_scope': pruning_scope,
            'pruning_frequency': pruning_frequency,
            'min_sparsity_ratio_per_op': min_sparsity_ratio_per_op,
            'max_sparsity_ratio_per_op': max_sparsity_ratio_per_op,
            'sparsity_decay_type': sparsity_decay_type,
            'pruning_op_types': pruning_op_types,
            'reg_type': reg_type,
            'criterion_reduce_type': criterion_reduce_type,
            'parameters': parameters,
            'resume_from_pruned_checkpoint': resume_from_pruned_checkpoint
        })


class WeightPruningConfig:
    """Similiar to torch optimizer's interface."""
    def __init__(self, pruning_configs=[{}],  ##empty dict will use global values
                 target_sparsity=0.9, pruning_type="snip_momentum", pattern="4x1", op_names=[],
                 excluded_op_names=[],
                 start_step=0, end_step=0, pruning_scope="global", pruning_frequency=1,
                 min_sparsity_ratio_per_op=0.0, max_sparsity_ratio_per_op=0.98,
                 sparsity_decay_type="exp", pruning_op_types=['Conv', 'Linear'],
                 **kwargs):
        """Init a WeightPruningConfig object."""
        self.pruning_configs = pruning_configs
        self._weight_compression = DotDict({
            'target_sparsity': target_sparsity,
            'pruning_type': pruning_type,
            'pattern': pattern,
            'op_names': op_names,
            'excluded_op_names': excluded_op_names,  ##global only
            'start_step': start_step,
            'end_step': end_step,
            'pruning_scope': pruning_scope,
            'pruning_frequency': pruning_frequency,
            'min_sparsity_ratio_per_op': min_sparsity_ratio_per_op,
            'max_sparsity_ratio_per_op': max_sparsity_ratio_per_op,
            'sparsity_decay_type': sparsity_decay_type,
            'pruning_op_types': pruning_op_types,
        })
        self._weight_compression.update(kwargs)

    @property
    def weight_compression(self):
        """Get weight_compression."""
        return self._weight_compression

    @weight_compression.setter
    def weight_compression(self, weight_compression):
        """Set weight_compression."""
        self._weight_compression = weight_compression



"""
This module is only used as reference to convert Python docstring to API document.

The created the API document is in `API Doc`_.

Usage::

    $python api_doc_example.py

Example::

    def fun(a):
        return a+1

    x = fun(2)
    print(x)

Attributes:
    attribute1 (int): Module level attribute.

        Please set 1-100 integer.

Todo:
    * Improve the code format
    * Add more comments

.. _API Doc:
   https://intel.github.io/intel-extension-for-transformers

"""

module_debug_level1 = 1

"""int: Module debug level document.

"""


def function1(param1, param2):
    """Example function1.

    Args:
        param1 (str): The parameter1.
        param2 (float): The parameter2.

    Example::

        >>> python api_doc_example.py
        >>> import os
        ... for i in range(3)
        ...     print(i)
        0
        1
        2

    Returns:
        bool: The return value. True|False.
    """


def function2(param1: str, param2: float) -> bool:
    """function with PEP 484 type annotations.

    Args:
        param1: The parameter1.
        param2: The parameter2.

    Example:

        Style 3::

            from neural_compressor.config import MixedPrecisionConfig
            def eval_func(model):
                ...
                return accuracy

            conf = MixedPrecisionConfig()
            output_model = mix_precision.fit(
                model,
                conf,
                eval_func=eval_func,
            )

    Returns:
        The return value. True|False.

    """


def function3(param1, param2=None, *args, **kwargs):
    """This is an example of function3.

    If ``*args`` or ``**kwargs`` are accepted,
    they should be listed as ``*args`` and ``**kwargs``.

    Args:
        param1 (int): The parameter1.
        param2 (:obj:`str`, optional): The parameter2.
        *args: Arguments list.
        **kwargs: Key-value dict.

    Returns:
        bool: The return value. True|False.

        The ``Returns`` section supports any reStructuredText formatting,
        including literal blocks::

            {
                'param1': param1,
                'param2': param2
            }

    Raises:
        AttributeError: The ``Raises`` section is a list of exceptions.
        ValueError: If `param2` is equal to `param1`.

    """
    if param1 == param2:
        raise ValueError('param1 may not be equal to param2')
    return True


def generator1(n):
    """Generators have a ``Yields`` section.

    Args:
        n (int): range.

    Yields:
        int: The next number in [0, `n` - 1].

    Examples::

        >>> print([i for i in example_generator(4)])
        [0, 1, 2, 3]

    """
    yield from range(n)


class ExampleClass:
    """Example for Class.

    If the class has public attributes, they may be documented here
    in an ``Attributes`` section.

    Attributes:
        attr1 (str): Description of `attr1`.
        attr2 (:obj:`int`, optional): Description of `attr2`.

    """

    def __init__(self, param1, param2, param3):
        """Example of __init__ method.

        Note:
            Do not include the `self` parameter in the ``Args`` section.

        Args:
            param1 (str): Description of `param1`.
            param2 (:obj:`int`, optional): Description of `param2`. Multiple
                lines are supported.
            param3 (list(str)): Description of `param3`.

        """
        self.attr1 = param1
        self.attr2 = param2
        self.attr3 = param3  #: Doc comment *inline*

        #: list(str): Doc comment *before* attribute, with type specified
        self.attr4 = ['attr4']

        self.attr5 = None
        """str: Docstring *after* attribute, with type specified."""

    @property
    def property1(self):
        """str: Property is documented."""
        return 'property1'


    def method1(self, param1, param2):
        """method1 for execute.

        Note:
            It's public.

        Args:
            param1: The parameter1.
            param2: The parameter2.

        Returns:
            True|False.

        """
        return True

    def __special__(self):
        """This function won't be documented that start with and
        end with a double underscore.
        """
        pass

    def _private(self):
        """private members are not included.
        """
        pass




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The QuantizedGraphDtypeRefactor Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from ..ops import Tensor
from .. import graph_utils as util
from .subgraph_matcher import EXECUTOR_TYPE
from .. import logger
import copy

@pattern_registry(pattern_type='Int8BF16MixedPrecisionChecker')
class Int8BF16MixedPrecisionChecker(Pattern):
    """The Int8BF16MixedPrecisionChecker pattern.

    Check if the model can be inferenced under in8/bf16 mixed precision. And modify graph if need.
    Quant -> InnerProduct / Matmul -> other Op
    u8 / s8 -> fp32 -> fp32 -----> u8 / s8 -> bf16 -> bf16
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        if model.framework_modeling_config['framework'] != 'torch':
            return model
        quant_info = util.get_quant_info()
        if not quant_info:
            return model
        def _get_dst_ops(node, model):
            ret = []
            output_name = node.output_tensors[0].name
            for node in model.nodes:
                for input_tensor in node.input_tensors:
                    if output_name == input_tensor.name:
                        ret.append(node)
            return ret

        def _all_dst_ops(node, model, checker, output_dtype):
            dst_ops = _get_dst_ops(node, model)
            for op in dst_ops:
                valid_connections = False if model.get_tensor_idx(op.name,
                                    node.output_tensors[0].name, from_output=False) == -1 else True
                o_type = EXECUTOR_TYPE.get(op.op_type, op.op_type)
                if not valid_connections or o_type in ['Shape', 'Quantize', 'Output', 'Softmax']:
                    continue
                if o_type not in checker[output_dtype] or \
                   op.attr.get('output_dtype', 'fp32') != output_dtype:
                    return False
            return True
        
        def _insert_bf16_quant_node(pre_node, model):
            output_tensor = copy.deepcopy(pre_node.output_tensors[0])
            output_tensor.dtype = 'bf16'
            pre_node.output_tensors[0].name = pre_node.output_tensors[0].name + '_before_quant'
            pre_node.output_tensors[0].dest_op = [pre_node.name + '_quant']
            input_tensor = copy.deepcopy(pre_node.output_tensors[0])
            quantize_op = util.construct_node(
                node_name=pre_node.name + "_quant",
                op_type='Quantize',
                input_tensors=[input_tensor],
                output_tensors=[output_tensor],
                attr=OrderedDict({'output_dtype': 'bf16'}))
            insert_idx = model.get_node_id(pre_node.name)
            model.insert_nodes(insert_idx + 1, [quantize_op])
            return model

        def _check_dst_op(start_node, model, checker, output_dtype):
            dst_ops = _get_dst_ops(start_node, model)
            if len(dst_ops) == 0:
                return
            for op in dst_ops:
                if EXECUTOR_TYPE.get(op.op_type, op.op_type) in checker[output_dtype]:
                    if op.attr.get('output_dtype', 'fp32') == 'fp32':
                        op.attr['output_dtype'] = output_dtype
                        _check_dst_op(op, model, checker, output_dtype)
                else:
                    continue

        def _scan_nodes_graph_dtype(model, checker, output_dtype):
            for node in model.nodes:
                o_type = EXECUTOR_TYPE.get(node.op_type, node.op_type)
                # int8 matrix multiply can output bf16 dtype but fp32 cannot
                if (o_type in ['InnerProduct', 'Matmul', 'Convolution',] and \
                    len(node.input_tensors) > 3):
                    if node.attr.get('output_dtype', 'fp32') == 'fp32':
                        node.attr['output_dtype'] = output_dtype
                        _check_dst_op(node, model, checker, output_dtype)

        def _fall_back_to_fp32(model, checker):
            for node in model.nodes:
                fall_back_node = None
                if EXECUTOR_TYPE.get(node.op_type, node.op_type) in ['Softmax'] and \
                   node.attr.get('output_dtype', 'fp32') in ['u8', 's8']:
                    pre_node = None
                    try:
                        pre_node = model.get_node_by_name(node.input_tensors[0].source_op[0])
                    except:
                        pre_node = None
                    if pre_node and pre_node.attr.get('output_dtype', 'fp32') == 'bf16':
                        fall_back_node = pre_node
                else:
                    if node.attr and node.attr.get('output_dtype', 'fp32') == 'bf16':
                        if not _all_dst_ops(node, model, checker, 'bf16'):
                            fall_back_node = node
                if fall_back_node:
                    if EXECUTOR_TYPE.get(fall_back_node.op_type, fall_back_node.op_type) in \
                       checker['bf16_fp32_mixed']:
                        fall_back_node.attr['output_dtype'] = 'fp32'
                    else:
                        logger.warning("{} can not fall back into fp32 dtype".format(
                                        fall_back_node.name))
                        return False
            return True

        def _revert_logits_output_dtype(model, output_dtype):
            for t in model.nodes[-1].input_tensors:
                try:
                    pre_node = model.get_node_by_name(t.source_op[0])
                except:
                    import pdb; pdb.set_trace()
                if pre_node and EXECUTOR_TYPE.get(pre_node.op_type, pre_node.op_type) in \
                   ['InnerProduct', 'Softmax', 'LogSoftmax']:
                    if pre_node.attr.get('output_dtype', "fp32") == output_dtype:
                        pre_node.attr['output_dtype'] = 'fp32'
        model_cpy = copy.deepcopy(model)
        # all in and out tensors support the dtype
        bf16_op = ['InnerProduct', 'Slice', 'Matmul', 'Reshape', 'BinaryOp',
                   'BinaryAdd', 'Reorder', 'Concat', 'Softmax', 'LayerNorm',
                   'LogSoftmax', 'Convolution', 'Gather']
        int8_op = ['InnerProduct', 'Matmul']
        bf16_fp32_mixed_op = ['InnerProduct', 'Matmul', 'BinaryOp', 'BinaryAdd', 'LayerNorm',
                              'Convolution', 'Softmax', 'Reorder', 'LogSoftmax']
        checker = {'bf16': bf16_op, 'int8': int8_op, 'bf16_fp32_mixed': bf16_fp32_mixed_op}
        # quant cos and sin nodes to bf16
        non_quantized_patterns = [[(0, 'Range'), (1, ['Div', 'BinaryOp']), (2, 'Pow'),
                                  (3, ['Div', 'BinaryOp']), (4, 'Reshape'),
                                  (7, ['Matmul', 'Einsum', 'BatchMatmul'])],
                                [(), (5, 'Range'), (6, 'Reshape'), 
                                 (7, ['Matmul', 'Einsum', 'BatchMatMul'])]]
        match_ret = util.search_pattern(non_quantized_patterns, model)
        for ret in match_ret:
            mat_node = model.get_node_by_name(ret[-2])
            dst_ops = _get_dst_ops(mat_node, model)
            dst_ops_type = [EXECUTOR_TYPE.get(n.op_type, n.op_type) for n in dst_ops]
            if dst_ops_type == ['CosSin', 'CosSin'] or dst_ops_type == ['Sin', 'Cos']:
                for n in dst_ops:
                    model = _insert_bf16_quant_node(n, model)

        for node in model.nodes:
            if EXECUTOR_TYPE.get(node.op_type, node.op_type) == 'Quantize' and \
               node.attr['output_dtype'] in ['u8', 's8', 'bf16']:
                dst_ops = _get_dst_ops(node, model)
                for op in dst_ops:
                    if op.attr.get('output_dtype', 'fp32') == 'fp32' and \
                        EXECUTOR_TYPE.get(op.op_type, op.op_type) in checker['bf16']:
                        op.attr['output_dtype'] = 'bf16'
                        _check_dst_op(op, model, checker, 'bf16')
        _scan_nodes_graph_dtype(model, checker, 'bf16')
        status = _fall_back_to_fp32(model, checker)
        _revert_logits_output_dtype(model, "bf16")
        for t in model.nodes[0].output_tensors:
            if t.location and len(t.location) == 2:
                break
            else:
                if t.dtype == "fp32":
                    t.dtype = "bf16"
        if status:
            del model_cpy
            return model
        else:
            del model
            logger.warning("The model cannot use int8/bf16 mixed precision...")
            return model_cpy
